import threading
from collections import deque
from typing import TYPE_CHECKING, List, Optional

try:
    import psutil
except ImportError:
    psutil = None
from .aggregators import aggregate_last, aggregate_mean
from .asset_registry import asset_registry
from .interfaces import Interface, Metric, MetricsMonitor

if TYPE_CHECKING:
    from typing import Deque

    from wandb.sdk.internal.settings_static import SettingsStatic


# CPU Metrics


class ProcessCpuPercent:
    """CPU usage of the process in percent normalized by the number of CPUs."""

    # name = "process_cpu_percent"
    name = "cpu"

    def __init__(self, pid: int) -> None:
        self.pid = pid
        self.samples: "Deque[float]" = deque([])
        self.process: Optional[psutil.Process] = None

    def sample(self) -> None:
        # todo: this is what we'd eventually want to do
        # self.samples.append(
        #     (
        #         datetime.datetime.utcnow(),
        #         psutil.Process(self.pid).cpu_percent(),
        #     )
        # )
        if self.process is None:
            self.process = psutil.Process(self.pid)

        self.samples.append(self.process.cpu_percent() / psutil.cpu_count())

    def clear(self) -> None:
        self.samples.clear()

    def aggregate(self) -> dict:
        # todo: create a statistics class with helper methods to compute
        #      mean, median, min, max, etc.
        if not self.samples:
            return {}
        aggregate = aggregate_mean(self.samples)
        return {self.name: aggregate}


class CpuPercent:
    """CPU usage of the system in percent per core."""

    name = "cpu.{i}.cpu_percent"

    def __init__(self, interval: Optional[float] = None) -> None:
        self.samples: "Deque[List[float]]" = deque([])
        self.interval = interval

    def sample(self) -> None:
        self.samples.append(psutil.cpu_percent(interval=self.interval, percpu=True))

    def clear(self) -> None:
        self.samples.clear()

    def aggregate(self) -> dict:
        if not self.samples:
            return {}
        num_cpu = len(self.samples[0])
        cpu_metrics = {}
        for i in range(num_cpu):
            aggregate_i = aggregate_mean([sample[i] for sample in self.samples])
            cpu_metrics[self.name.format(i=i)] = aggregate_i

        return cpu_metrics


class ProcessCpuThreads:
    """Number of threads used by the process."""

    name = "proc.cpu.threads"

    def __init__(self, pid: int) -> None:
        self.samples: "Deque[int]" = deque([])
        self.pid = pid
        self.process: Optional[psutil.Process] = None

    def sample(self) -> None:
        if self.process is None:
            self.process = psutil.Process(self.pid)

        self.samples.append(self.process.num_threads())

    def clear(self) -> None:
        self.samples.clear()

    def aggregate(self) -> dict:
        if not self.samples:
            return {}
        return {self.name: aggregate_last(self.samples)}


@asset_registry.register
class CPU:
    def __init__(
        self,
        interface: "Interface",
        settings: "SettingsStatic",
        shutdown_event: threading.Event,
    ) -> None:
        self.name: str = self.__class__.__name__.lower()
        self.metrics: List[Metric] = [
            ProcessCpuPercent(settings._stats_pid),
            CpuPercent(),
            ProcessCpuThreads(settings._stats_pid),
        ]
        self.metrics_monitor: "MetricsMonitor" = MetricsMonitor(
            self.name,
            self.metrics,
            interface,
            settings,
            shutdown_event,
        )

    @classmethod
    def is_available(cls) -> bool:
        return psutil is not None

    def probe(self) -> dict:
        asset_info = {
            "cpu_count": psutil.cpu_count(logical=False),
            "cpu_count_logical": psutil.cpu_count(logical=True),
        }
        try:
            asset_info["cpu_freq"] = {
                "current": psutil.cpu_freq().current,
                "min": psutil.cpu_freq().min,
                "max": psutil.cpu_freq().max,
            }
            asset_info["cpu_freq_per_core"] = [
                {
                    "current": freq.current,
                    "min": freq.min,
                    "max": freq.max,
                }
                for freq in psutil.cpu_freq(percpu=True)
            ]
        except Exception:
            pass
        return asset_info

    def start(self) -> None:
        self.metrics_monitor.start()

    def finish(self) -> None:
        self.metrics_monitor.finish()



"""Calculates and formats metrics and charts for introspecting sklearn models.

The functions in these modules are designed to be called by functions from the
plot submodule that have been exported into the namespace of the wandb.sklearn
submodule, rather than being called directly.
"""

from .calibration_curves import calibration_curves
from .class_proportions import class_proportions
from .confusion_matrix import confusion_matrix
from .decision_boundaries import decision_boundaries
from .elbow_curve import elbow_curve
from .feature_importances import feature_importances
from .learning_curve import learning_curve
from .outlier_candidates import outlier_candidates
from .residuals import residuals
from .silhouette import silhouette
from .summary_metrics import summary_metrics

__all__ = [
    "calibration_curves",
    "class_proportions",
    "confusion_matrix",
    "decision_boundaries",
    "elbow_curve",
    "feature_importances",
    "learning_curve",
    "outlier_candidates",
    "residuals",
    "silhouette",
    "summary_metrics",
]



from typing import Any, Dict, Iterable, Optional

from wandb.util import get_module

from .base import Importer, ImporterRun

mlflow = get_module(
    "mlflow",
    required="To use the MlflowImporter, please install mlflow: `pip install mlflow`",
)


class MlflowRun(ImporterRun):
    def __init__(self, run, mlflow_client):
        self.run = run
        self.mlflow_client = mlflow_client
        super().__init__()

    def run_id(self):
        return self.run.info.run_id

    def entity(self):
        return self.run.info.user_id

    def project(self):
        return "imported-from-mlflow"

    def config(self):
        return self.run.data.params

    def summary(self):
        return self.run.data.metrics

    def metrics(self):
        def wandbify(metrics):
            for step, t in enumerate(metrics):
                d = {m.key: m.value for m in t}
                d["_step"] = step
                yield d

        metrics = [
            self.mlflow_client.get_metric_history(self.run.info.run_id, k)
            for k in self.run.data.metrics.keys()
        ]
        metrics = zip(*metrics)  # transpose
        return wandbify(metrics)

        # Alternate: Might be slower but use less mem
        # Can't make this a generator.  See mlflow get_metric_history internals
        # https://github.com/mlflow/mlflow/blob/master/mlflow/tracking/_tracking_service/client.py#L74-L93
        # for k in self.run.data.metrics.keys():
        #     history = self.mlflow_client.get_metric_history(self.run.info.run_id, k)
        #     yield wandbify(history)

    def run_group(self):
        # this is nesting?  Parent at `run.info.tags.get("mlflow.parentRunId")`
        return f"Experiment {self.run.info.experiment_id}"

    def job_type(self):
        # Is this the right approach?
        return f"User {self.run.info.user_id}"

    def display_name(self):
        return self.run.info.run_name

    def notes(self):
        return self.run.data.tags.get("mlflow.note.content")

    def tags(self):
        return {
            k: v for k, v in self.run.data.tags.items() if not k.startswith("mlflow.")
        }

    def start_time(self):
        return self.run.info.start_time // 1000

    def runtime(self):
        return self.run.info.end_time // 1_000 - self.start_time()

    def git(self):
        ...

    def artifacts(self):
        for f in self.mlflow_client.list_artifacts(self.run.info.run_id):
            dir_path = mlflow.artifacts.download_artifacts(run_id=self.run.info.run_id)
            full_path = dir_path + f.path
            yield (f.path, full_path)


class MlflowImporter(Importer):
    def __init__(
        self, mlflow_tracking_uri, mlflow_registry_uri=None, wandb_base_url=None
    ) -> None:
        super().__init__()
        self.mlflow_tracking_uri = mlflow_tracking_uri

        mlflow.set_tracking_uri(self.mlflow_tracking_uri)
        if mlflow_registry_uri:
            mlflow.set_registry_uri(mlflow_registry_uri)
        self.mlflow_client = mlflow.tracking.MlflowClient(mlflow_tracking_uri)

    def import_one(
        self,
        run: ImporterRun,
        overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        mlflow.set_tracking_uri(self.mlflow_tracking_uri)
        super().import_one(run, overrides)

    def download_all_runs(self) -> Iterable[MlflowRun]:
        for exp in self.mlflow_client.search_experiments():
            for run in self.mlflow_client.search_runs(exp.experiment_id):
                yield MlflowRun(run, self.mlflow_client)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011 Yesudeep Mangalapilly <yesudeep@gmail.com>
# Copyright 2012 Google, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:module: watchdog.observers.fsevents
:synopsis: FSEvents based emitter implementation.
:author: yesudeep@google.com (Yesudeep Mangalapilly)
:platforms: Mac OS X
"""

from __future__ import with_statement

import sys
import threading
import unicodedata
import _watchdog_fsevents as _fsevents

from wandb_watchdog.events import (
    FileDeletedEvent,
    FileModifiedEvent,
    FileCreatedEvent,
    FileMovedEvent,
    DirDeletedEvent,
    DirModifiedEvent,
    DirCreatedEvent,
    DirMovedEvent
)

from wandb_watchdog.utils.dirsnapshot import DirectorySnapshot
from wandb_watchdog.observers.api import (
    BaseObserver,
    EventEmitter,
    DEFAULT_EMITTER_TIMEOUT,
    DEFAULT_OBSERVER_TIMEOUT
)


class FSEventsEmitter(EventEmitter):

    """
    Mac OS X FSEvents Emitter class.

    :param event_queue:
        The event queue to fill with events.
    :param watch:
        A watch object representing the directory to monitor.
    :type watch:
        :class:`watchdog.observers.api.ObservedWatch`
    :param timeout:
        Read events blocking timeout (in seconds).
    :type timeout:
        ``float``
    """

    def __init__(self, event_queue, watch, timeout=DEFAULT_EMITTER_TIMEOUT):
        EventEmitter.__init__(self, event_queue, watch, timeout)
        self._lock = threading.Lock()
        self.snapshot = DirectorySnapshot(watch.path, watch.is_recursive)

    def on_thread_stop(self):
        _fsevents.remove_watch(self.watch)
        _fsevents.stop(self)

    def queue_events(self, timeout):
        with self._lock:
            if not self.watch.is_recursive\
                and self.watch.path not in self.pathnames:
                return
            new_snapshot = DirectorySnapshot(self.watch.path,
                                             self.watch.is_recursive)
            events = new_snapshot - self.snapshot
            self.snapshot = new_snapshot

            # Files.
            for src_path in events.files_deleted:
                self.queue_event(FileDeletedEvent(src_path))
            for src_path in events.files_modified:
                self.queue_event(FileModifiedEvent(src_path))
            for src_path in events.files_created:
                self.queue_event(FileCreatedEvent(src_path))
            for src_path, dest_path in events.files_moved:
                self.queue_event(FileMovedEvent(src_path, dest_path))

            # Directories.
            for src_path in events.dirs_deleted:
                self.queue_event(DirDeletedEvent(src_path))
            for src_path in events.dirs_modified:
                self.queue_event(DirModifiedEvent(src_path))
            for src_path in events.dirs_created:
                self.queue_event(DirCreatedEvent(src_path))
            for src_path, dest_path in events.dirs_moved:
                self.queue_event(DirMovedEvent(src_path, dest_path))

    def run(self):
        try:
            def callback(pathnames, flags, emitter=self):
                emitter.queue_events(emitter.timeout)

            # for pathname, flag in zip(pathnames, flags):
            # if emitter.watch.is_recursive: # and pathname != emitter.watch.path:
            #    new_sub_snapshot = DirectorySnapshot(pathname, True)
            #    old_sub_snapshot = self.snapshot.copy(pathname)
            #    diff = new_sub_snapshot - old_sub_snapshot
            #    self.snapshot += new_subsnapshot
            # else:
            #    new_snapshot = DirectorySnapshot(emitter.watch.path, False)
            #    diff = new_snapshot - emitter.snapshot
            #    emitter.snapshot = new_snapshot

            # INFO: FSEvents reports directory notifications recursively
            # by default, so we do not need to add subdirectory paths.
            #pathnames = set([self.watch.path])
            # if self.watch.is_recursive:
            #    for root, directory_names, _ in os.walk(self.watch.path):
            #        for directory_name in directory_names:
            #            full_path = absolute_path(
            #                            os.path.join(root, directory_name))
            #            pathnames.add(full_path)
            self.pathnames = [self.watch.path]
            _fsevents.add_watch(self,
                                self.watch,
                                callback,
                                self.pathnames)
            _fsevents.read_events(self)
        except:
            pass


class FSEventsObserver(BaseObserver):

    def __init__(self, timeout=DEFAULT_OBSERVER_TIMEOUT):
        BaseObserver.__init__(self, emitter_class=FSEventsEmitter,
                              timeout=timeout)

    def schedule(self, event_handler, path, recursive=False):
        # Python 2/3 compat
        try:
            str_class = unicode
        except NameError:
            str_class = str

        # Fix for issue #26: Trace/BPT error when given a unicode path
        # string. https://github.com/gorakhargosh/watchdog/issues#issue/26
        if isinstance(path, str_class):
            #path = unicode(path, 'utf-8')
            path = unicodedata.normalize('NFC', path)
            # We only encode the path in Python 2 for backwards compatibility.
            # On Python 3 we want the path to stay as unicode if possible for
            # the sake of path matching not having to be rewritten to use the
            # bytes API instead of strings. The _watchdog_fsevent.so code for
            # Python 3 can handle both str and bytes paths, which is why we
            # do not HAVE to encode it with Python 3. The Python 2 code in
            # _watchdog_fsevents.so was not changed for the sake of backwards
            # compatibility.
            if sys.version_info < (3,):
                path = path.encode('utf-8')
        return BaseObserver.schedule(self, event_handler, path, recursive)



# flake8: noqa
from inspect import cleandoc

from ... import termlog
from . import blocks, helpers, panels, templates
from .blocks import *
from .helpers import LineKey, PCColumn
from .panels import *
from .report import Report
from .runset import Runset
from .templates import *
from .util import InlineCode, InlineLaTeX, Link

termlog(
    cleandoc(
        """
        Thanks for trying out the Report API!
        For a tutorial, check out https://colab.research.google.com/drive/1CzyJx1nuOS4pdkXa2XPaRQyZdmFmLmXV

        Try out tab completion to see what's available.
          ∟ everything:    `wr.<tab>`
              ∟ panels:    `wr.panels.<tab>`
              ∟ blocks:    `wr.blocks.<tab>`
              ∟ helpers:   `wr.helpers.<tab>`
              ∟ templates: `wr.templates.<tab>`
              
        For bugs/feature requests, please create an issue on github: https://github.com/wandb/wandb/issues
        """
    )
)



import functools
import logging
import sys
from typing import Any, Dict, List, Optional

import wandb.sdk
import wandb.util
from wandb.sdk.lib import telemetry as wb_telemetry
from wandb.sdk.lib.timer import Timer

from .resolver import OpenAIRequestResponseResolver

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


logger = logging.getLogger(__name__)


AutologOpenAIInitArgs = Optional[Dict[str, Any]]


class PatchOpenAIAPI:
    symbols: List[Literal["Edit", "Completion", "ChatCompletion"]] = [
        "Edit",
        "Completion",
        "ChatCompletion",
    ]

    def __init__(self) -> None:
        """Patches the OpenAI API to log traces to W&B."""
        self.original_methods: Dict[str, Any] = {}
        self.resolver = OpenAIRequestResponseResolver()
        self._openai = None

    @property
    def openai(self) -> Any:
        """Returns the openai module."""
        if self._openai is None:
            self._openai = wandb.util.get_module(
                name="openai",
                required="To use the W&B OpenAI Autolog, you need to have the `openai` python "
                "package installed. Please install it with `pip install openai`.",
                lazy=False,
            )
        return self._openai

    def patch(self, run: "wandb.sdk.wandb_run.Run") -> None:
        """Patches the OpenAI API to log traces to W&B."""
        for symbol in self.symbols:
            original = getattr(self.openai, symbol).create

            def method_factory(original_method: Any):
                @functools.wraps(original_method)
                def create(*args, **kwargs):
                    with Timer() as timer:
                        result = original_method(*args, **kwargs)
                    try:
                        trace = self.resolver(kwargs, result, timer.elapsed)
                        if trace is not None:
                            run.log({"trace": trace})
                    except Exception:
                        # logger.warning(e)
                        pass
                    return result

                return create

            # save original method
            self.original_methods[symbol] = original
            # monkeypatch
            getattr(self.openai, symbol).create = method_factory(original)

    def unpatch(self) -> None:
        """Unpatches the OpenAI API."""
        for symbol, original in self.original_methods.items():
            getattr(self.openai, symbol).create = original


class AutologOpenAI:
    def __init__(self) -> None:
        """Autolog OpenAI API calls to W&B."""
        self._patch_openai_api = PatchOpenAIAPI()
        self._run: Optional["wandb.sdk.wandb_run.Run"] = None
        self.__run_created_by_autolog: bool = False

    @property
    def _is_enabled(self) -> bool:
        """Returns whether autologging is enabled."""
        return self._run is not None

    def __call__(self, init: AutologOpenAIInitArgs = None) -> None:
        """Enable OpenAI autologging."""
        self.enable(init=init)

    def _run_init(self, init: AutologOpenAIInitArgs = None) -> None:
        """Handle wandb run initialization."""
        # - autolog(init: dict = {...}) calls wandb.init(**{...})
        #   regardless of whether there is a wandb.run or not,
        #   we only track if the run was created by autolog
        #    - todo: autolog(init: dict | run = run) would use the user-provided run
        # - autolog() uses the wandb.run if there is one, otherwise it calls wandb.init()
        if init:
            _wandb_run = wandb.run
            # we delegate dealing with the init dict to wandb.init()
            self._run = wandb.init(**init)
            if _wandb_run != self._run:
                self.__run_created_by_autolog = True
        elif wandb.run is None:
            self._run = wandb.init()
            self.__run_created_by_autolog = True
        else:
            self._run = wandb.run

    def enable(self, init: AutologOpenAIInitArgs = None) -> None:
        """Enable OpenAI autologging.

        Args:
            init: Optional dictionary of arguments to pass to wandb.init().

        """
        if self._is_enabled:
            logger.info(
                "OpenAI autologging is already enabled, disabling and re-enabling."
            )
            self.disable()

        logger.info("Enabling OpenAI autologging.")
        self._run_init(init=init)

        self._patch_openai_api.patch(self._run)

        with wb_telemetry.context(self._run) as tel:
            tel.feature.openai_autolog = True

    def disable(self) -> None:
        """Disable OpenAI autologging."""
        if self._run is None:
            return

        logger.info("Disabling OpenAI autologging.")

        if self.__run_created_by_autolog:
            self._run.finish()
            self.__run_created_by_autolog = False

        self._run = None

        self._patch_openai_api.unpatch()



try:
    # Try to load the Cython performant OrderedDict (C)
    # as is more performant than collections.OrderedDict (Python)
    from cyordereddict import OrderedDict
except ImportError:
    from collections import OrderedDict

__all__ = ['OrderedDict']



# WARNING: This is an anti-pattern file and we should avoid
# adding to it and remove entries whenever possible. This file
# contains global objects which need to be referenced by multiple
# submodules. If you need a global object, seriously reconsider. This
# file is intended to be a stop gap to help during code migrations (eg.
# when moving to typing a module) to avoid circular references. Anything
# added here is pure tech debt. Use with care. - Tim

_glob_datatypes_callback = None


def _datatypes_set_callback(cb):
    global _glob_datatypes_callback
    _glob_datatypes_callback = cb


def _datatypes_callback(fname):
    if _glob_datatypes_callback:
        _glob_datatypes_callback(fname)



import json

from ..error import GraphQLSyntaxError

__all__ = ['Token', 'Lexer', 'TokenKind',
           'get_token_desc', 'get_token_kind_desc']


class Token(object):
    __slots__ = 'kind', 'start', 'end', 'value'

    def __init__(self, kind, start, end, value=None):
        self.kind = kind
        self.start = start
        self.end = end
        self.value = value

    def __repr__(self):
        return u'<Token kind={} at {}..{} value={}>'.format(
            get_token_kind_desc(self.kind),
            self.start,
            self.end,
            repr(self.value)
        )

    def __eq__(self, other):
        return (self.kind == other.kind and
                self.start == other.start and
                self.end == other.end and
                self.value == other.value)


class Lexer(object):
    __slots__ = 'source', 'prev_position'

    def __init__(self, source):
        self.source = source
        self.prev_position = 0

    def next_token(self, reset_position=None):
        if reset_position is None:
            reset_position = self.prev_position
        token = read_token(self.source, reset_position)
        self.prev_position = token.end
        return token


class TokenKind(object):
    EOF = 1
    BANG = 2
    DOLLAR = 3
    PAREN_L = 4
    PAREN_R = 5
    SPREAD = 6
    COLON = 7
    EQUALS = 8
    AT = 9
    BRACKET_L = 10
    BRACKET_R = 11
    BRACE_L = 12
    PIPE = 13
    BRACE_R = 14
    NAME = 15
    VARIABLE = 16
    INT = 17
    FLOAT = 18
    STRING = 19


def get_token_desc(token):
    if token.value:
        return u'{} "{}"'.format(
            get_token_kind_desc(token.kind),
            token.value
        )
    else:
        return get_token_kind_desc(token.kind)


def get_token_kind_desc(kind):
    return TOKEN_DESCRIPTION[kind]


TOKEN_DESCRIPTION = {
    TokenKind.EOF: 'EOF',
    TokenKind.BANG: '!',
    TokenKind.DOLLAR: '$',
    TokenKind.PAREN_L: '(',
    TokenKind.PAREN_R: ')',
    TokenKind.SPREAD: '...',
    TokenKind.COLON: ':',
    TokenKind.EQUALS: '=',
    TokenKind.AT: '@',
    TokenKind.BRACKET_L: '[',
    TokenKind.BRACKET_R: ']',
    TokenKind.BRACE_L: '{',
    TokenKind.PIPE: '|',
    TokenKind.BRACE_R: '}',
    TokenKind.NAME: 'Name',
    TokenKind.VARIABLE: 'Variable',
    TokenKind.INT: 'Int',
    TokenKind.FLOAT: 'Float',
    TokenKind.STRING: 'String',
}


def char_code_at(s, pos):
    if 0 <= pos < len(s):
        return ord(s[pos])

    return None


PUNCT_CODE_TO_KIND = {
    ord('!'): TokenKind.BANG,
    ord('$'): TokenKind.DOLLAR,
    ord('('): TokenKind.PAREN_L,
    ord(')'): TokenKind.PAREN_R,
    ord(':'): TokenKind.COLON,
    ord('='): TokenKind.EQUALS,
    ord('@'): TokenKind.AT,
    ord('['): TokenKind.BRACKET_L,
    ord(']'): TokenKind.BRACKET_R,
    ord('{'): TokenKind.BRACE_L,
    ord('|'): TokenKind.PIPE,
    ord('}'): TokenKind.BRACE_R,
}


def print_char_code(code):
    if code is None:
        return '<EOF>'

    if code < 0x007F:
        return json.dumps(chr(code))

    return '"\\u%04X"' % code


def read_token(source, from_position):
    """Gets the next token from the source starting at the given position.

    This skips over whitespace and comments until it finds the next lexable
    token, then lexes punctuators immediately or calls the appropriate
    helper fucntion for more complicated tokens."""
    body = source.body
    body_length = len(body)

    position = position_after_whitespace(body, from_position)

    if position >= body_length:
        return Token(TokenKind.EOF, position, position)

    code = char_code_at(body, position)

    if code < 0x0020 and code not in (0x0009, 0x000A, 0x000D):
        raise GraphQLSyntaxError(
            source, position,
            u'Invalid character {}.'.format(print_char_code(code))
        )

    kind = PUNCT_CODE_TO_KIND.get(code)
    if kind is not None:
        return Token(kind, position, position + 1)

    if code == 46:  # .
        if char_code_at(body, position + 1) == char_code_at(body, position + 2) == 46:
            return Token(TokenKind.SPREAD, position, position + 3)

    elif 65 <= code <= 90 or code == 95 or 97 <= code <= 122:
        # A-Z, _, a-z
        return read_name(source, position)

    elif code == 45 or 48 <= code <= 57:  # -, 0-9
        return read_number(source, position, code)

    elif code == 34:  # "
        return read_string(source, position)

    raise GraphQLSyntaxError(
        source, position,
        u'Unexpected character {}.'.format(print_char_code(code)))


ignored_whitespace_characters = frozenset([
    # BOM
    0xFEFF,
    # White Space
    0x0009,  # tab
    0x0020,  # space
    # Line Terminator
    0x000A,  # new line
    0x000D,  # carriage return
    # Comma
    0x002C
])


def position_after_whitespace(body, start_position):
    """Reads from body starting at start_position until it finds a
    non-whitespace or commented character, then returns the position of
    that character for lexing."""
    body_length = len(body)
    position = start_position
    while position < body_length:
        code = char_code_at(body, position)
        if code in ignored_whitespace_characters:
            position += 1

        elif code == 35:  # #, skip comments
            position += 1
            while position < body_length:
                code = char_code_at(body, position)
                if not (code is not None and (code > 0x001F or code == 0x0009) and code not in (0x000A, 0x000D)):
                    break

                position += 1
        else:
            break
    return position


def read_number(source, start, first_code):
    """Reads a number token from the source file, either a float
    or an int depending on whether a decimal point appears.

    Int:   -?(0|[1-9][0-9]*)
    Float: -?(0|[1-9][0-9]*)(\.[0-9]+)?((E|e)(+|-)?[0-9]+)?"""
    code = first_code
    body = source.body
    position = start
    is_float = False

    if code == 45:  # -
        position += 1
        code = char_code_at(body, position)

    if code == 48:  # 0
        position += 1
        code = char_code_at(body, position)

        if code is not None and 48 <= code <= 57:
            raise GraphQLSyntaxError(
                source,
                position,
                u'Invalid number, unexpected digit after 0: {}.'.format(print_char_code(code))
            )
    else:
        position = read_digits(source, position, code)
        code = char_code_at(body, position)

    if code == 46:  # .
        is_float = True

        position += 1
        code = char_code_at(body, position)
        position = read_digits(source, position, code)
        code = char_code_at(body, position)

    if code in (69, 101):  # E e
        is_float = True
        position += 1
        code = char_code_at(body, position)
        if code in (43, 45):  # + -
            position += 1
            code = char_code_at(body, position)

        position = read_digits(source, position, code)

    return Token(
        TokenKind.FLOAT if is_float else TokenKind.INT,
        start,
        position,
        body[start:position]
    )


def read_digits(source, start, first_code):
    body = source.body
    position = start
    code = first_code

    if code is not None and 48 <= code <= 57:  # 0 - 9
        while True:
            position += 1
            code = char_code_at(body, position)

            if not (code is not None and 48 <= code <= 57):
                break

        return position

    raise GraphQLSyntaxError(
        source,
        position,
        u'Invalid number, expected digit but got: {}.'.format(print_char_code(code))
    )


ESCAPED_CHAR_CODES = {
    34: '"',
    47: '/',
    92: '\\',
    98: '\b',
    102: '\f',
    110: '\n',
    114: '\r',
    116: '\t',
}


def read_string(source, start):
    """Reads a string token from the source file.

    "([^"\\\u000A\u000D\u2028\u2029]|(\\(u[0-9a-fA-F]{4}|["\\/bfnrt])))*"
    """
    body = source.body
    body_length = len(body)

    position = start + 1
    chunk_start = position
    code = 0
    value = []
    append = value.append

    while position < body_length:
        code = char_code_at(body, position)
        if not (
            code is not None and
            code not in (
                # LineTerminator
                0x000A, 0x000D,
                # Quote
                34
            )
        ):
            break

        if code < 0x0020 and code != 0x0009:
            raise GraphQLSyntaxError(
                source,
                position,
                u'Invalid character within String: {}.'.format(print_char_code(code))
            )

        position += 1
        if code == 92:  # \
            append(body[chunk_start:position - 1])

            code = char_code_at(body, position)
            escaped = ESCAPED_CHAR_CODES.get(code)
            if escaped is not None:
                append(escaped)

            elif code == 117:  # u
                char_code = uni_char_code(
                    char_code_at(body, position + 1) or 0,
                    char_code_at(body, position + 2) or 0,
                    char_code_at(body, position + 3) or 0,
                    char_code_at(body, position + 4) or 0,
                )

                if char_code < 0:
                    raise GraphQLSyntaxError(
                        source, position,
                        u'Invalid character escape sequence: \\u{}.'.format(body[position + 1: position + 5])
                    )

                append(chr(char_code))
                position += 4
            else:
                raise GraphQLSyntaxError(
                    source, position,
                    u'Invalid character escape sequence: \\{}.'.format(chr(code))
                )

            position += 1
            chunk_start = position

    if code != 34:  # Quote (")
        raise GraphQLSyntaxError(source, position, 'Unterminated string')

    append(body[chunk_start:position])
    return Token(TokenKind.STRING, start, position + 1, u''.join(value))


def uni_char_code(a, b, c, d):
    """Converts four hexidecimal chars to the integer that the
    string represents. For example, uniCharCode('0','0','0','f')
    will return 15, and uniCharCode('0','0','f','f') returns 255.

    Returns a negative number on error, if a char was invalid.

    This is implemented by noting that char2hex() returns -1 on error,
    which means the result of ORing the char2hex() will also be negative.
    """
    return (char2hex(a) << 12 | char2hex(b) << 8 |
            char2hex(c) << 4 | char2hex(d))


def char2hex(a):
    """Converts a hex character to its integer value.
    '0' becomes 0, '9' becomes 9
    'A' becomes 10, 'F' becomes 15
    'a' becomes 10, 'f' becomes 15

    Returns -1 on error."""
    if 48 <= a <= 57:  # 0-9
        return a - 48
    elif 65 <= a <= 70:  # A-F
        return a - 55
    elif 97 <= a <= 102:  # a-f
        return a - 87
    return -1


def read_name(source, position):
    """Reads an alphanumeric + underscore name from the source.

    [_A-Za-z][_0-9A-Za-z]*"""
    body = source.body
    body_length = len(body)
    end = position + 1

    while end != body_length:
        code = char_code_at(body, end)
        if not (code is not None and (
            code == 95 or  # _
            48 <= code <= 57 or  # 0-9
            65 <= code <= 90 or  # A-Z
            97 <= code <= 122  # a-z
        )):
            break

        end += 1

    return Token(TokenKind.NAME, position, end, body[position:end])



import codecs
import os
from typing import TYPE_CHECKING, Type, Union

from wandb import util
from wandb.sdk.lib import runid

from .._private import MEDIA_TMP
from .media import Media

if TYPE_CHECKING:  # pragma: no cover
    from ...wandb_artifacts import Artifact as LocalArtifact
    from ...wandb_run import Run as LocalRun


# Allows encoding of arbitrary JSON structures
# as a file
#
# This class should be used as an abstract class
# extended to have validation methods


class JSONMetadata(Media):
    """JSONMetadata is a type for encoding arbitrary metadata as files."""

    def __init__(self, val: dict) -> None:
        super().__init__()

        self.validate(val)
        self._val = val

        ext = "." + self.type_name() + ".json"
        tmp_path = os.path.join(MEDIA_TMP.name, runid.generate_id() + ext)
        with codecs.open(tmp_path, "w", encoding="utf-8") as fp:
            util.json_dump_uncompressed(self._val, fp)
        self._set_file(tmp_path, is_tmp=True, extension=ext)

    @classmethod
    def get_media_subdir(cls: Type["JSONMetadata"]) -> str:
        return os.path.join("media", "metadata", cls.type_name())

    def to_json(self, run_or_artifact: Union["LocalRun", "LocalArtifact"]) -> dict:
        json_dict = super().to_json(run_or_artifact)
        json_dict["_type"] = self.type_name()

        return json_dict

    # These methods should be overridden in the child class
    @classmethod
    def type_name(cls) -> str:
        return "metadata"

    def validate(self, val: dict) -> bool:
        return True



from pytest import mark
from promise import Promise
from promise.dataloader import DataLoader


def id_loader(**options):
    load_calls = []

    resolve = options.pop("resolve", Promise.resolve)

    def fn(keys):
        load_calls.append(keys)
        return resolve(keys)

    identity_loader = DataLoader(fn, **options)
    return identity_loader, load_calls


@mark.asyncio
async def test_await_dataloader():
    identity_loader, load_calls = id_loader()

    async def load_multiple(identity_loader):
        one = identity_loader.load("load1")
        two = identity_loader.load("load2")
        return await Promise.all([one, two])

    result = await load_multiple(identity_loader)
    assert result == ["load1", "load2"]
    assert load_calls == [["load1"], ["load2"]]


@mark.asyncio
async def test_await_dataloader_safe_promise():
    identity_loader, load_calls = id_loader()

    @Promise.safe
    async def load_multiple(identity_loader):
        one = identity_loader.load("load1")
        two = identity_loader.load("load2")
        return await Promise.all([one, two])

    result = await load_multiple(identity_loader)
    assert result == ["load1", "load2"]
    assert load_calls == [["load1"], ["load2"]]


@mark.asyncio
async def test_await_dataloader_individual():
    identity_loader, load_calls = id_loader()

    async def load_one_then_two(identity_loader):
        one = await identity_loader.load("load1")
        two = await identity_loader.load("load2")
        return [one, two]

    result = await load_one_then_two(identity_loader)
    assert result == ["load1", "load2"]
    assert load_calls == [["load1"], ["load2"]]


@mark.asyncio
async def test_await_dataloader_individual_safe_promise():
    identity_loader, load_calls = id_loader()

    @Promise.safe
    async def load_one_then_two(identity_loader):
        one = await identity_loader.load("load1")
        two = await identity_loader.load("load2")
        return [one, two]

    result = await load_one_then_two(identity_loader)
    assert result == ["load1", "load2"]
    assert load_calls == [["load1"], ["load2"]]


@mark.asyncio
async def test_await_dataloader_two():
    identity_loader, load_calls = id_loader()

    async def load_one_then_two(identity_loader):
        one = await identity_loader.load("load1")
        two = await identity_loader.load("load2")
        return (one, two)

    result12 = await Promise.all([load_one_then_two(identity_loader)])


@mark.asyncio
async def test_await_dataloader_two_safe_promise():
    identity_loader, load_calls = id_loader()

    @Promise.safe
    async def load_one_then_two(identity_loader):
        one = await identity_loader.load("load1")
        two = await identity_loader.load("load2")
        return (one, two)

    result12 = await Promise.all([load_one_then_two(identity_loader)])



from ..language.parser import parse_value
from ..pyutils.ordereddict import OrderedDict
from ..type import (GraphQLArgument, GraphQLBoolean, GraphQLEnumType,
                    GraphQLEnumValue, GraphQLField, GraphQLFloat, GraphQLID,
                    GraphQLInputObjectField, GraphQLInputObjectType,
                    GraphQLInt, GraphQLInterfaceType, GraphQLList,
                    GraphQLNonNull, GraphQLObjectType, GraphQLScalarType,
                    GraphQLSchema, GraphQLString, GraphQLUnionType,
                    is_input_type, is_output_type)
from ..type.directives import DirectiveLocation, GraphQLDirective
from ..type.introspection import (TypeKind, __Directive, __DirectiveLocation,
                                  __EnumValue, __Field, __InputValue, __Schema,
                                  __Type, __TypeKind)
from .value_from_ast import value_from_ast


def _false(*_):
    return False


def _none(*_):
    return None


def no_execution(*args):
    raise Exception('Client Schema cannot be used for execution.')


def build_client_schema(introspection):
    schema_introspection = introspection['__schema']

    type_introspection_map = {t['name']: t for t in schema_introspection['types']}

    type_def_cache = {
        'String': GraphQLString,
        'Int': GraphQLInt,
        'Float': GraphQLFloat,
        'Boolean': GraphQLBoolean,
        'ID': GraphQLID,
        '__Schema': __Schema,
        '__Directive': __Directive,
        '__DirectiveLocation': __DirectiveLocation,
        '__Type': __Type,
        '__Field': __Field,
        '__InputValue': __InputValue,
        '__EnumValue': __EnumValue,
        '__TypeKind': __TypeKind,

    }

    def get_type(type_ref):
        kind = type_ref.get('kind')

        if kind == TypeKind.LIST:
            item_ref = type_ref.get('ofType')

            if not item_ref:
                raise Exception('Decorated type deeper than introspection query.')

            return GraphQLList(get_type(item_ref))

        elif kind == TypeKind.NON_NULL:
            nullable_ref = type_ref.get('ofType')
            if not nullable_ref:
                raise Exception('Decorated type deeper than introspection query.')

            return GraphQLNonNull(get_type(nullable_ref))

        return get_named_type(type_ref['name'])

    def get_named_type(type_name):
        if type_name in type_def_cache:
            return type_def_cache[type_name]

        type_introspection = type_introspection_map.get(type_name)
        if not type_introspection:
            raise Exception(
                'Invalid or incomplete schema, unknown type: {}. Ensure that a full introspection query '
                'is used in order to build a client schema.'.format(type_name)
            )

        type_def = type_def_cache[type_name] = build_type(type_introspection)
        return type_def

    def get_input_type(type_ref):
        input_type = get_type(type_ref)
        assert is_input_type(input_type), 'Introspection must provide input type for arguments.'
        return input_type

    def get_output_type(type_ref):
        output_type = get_type(type_ref)
        assert is_output_type(output_type), 'Introspection must provide output type for fields.'
        return output_type

    def get_object_type(type_ref):
        object_type = get_type(type_ref)
        assert isinstance(object_type, GraphQLObjectType), 'Introspection must provide object type for possibleTypes.'
        return object_type

    def get_interface_type(type_ref):
        interface_type = get_type(type_ref)
        assert isinstance(interface_type, GraphQLInterfaceType), \
            'Introspection must provide interface type for interfaces.'
        return interface_type

    def build_type(type):
        type_kind = type.get('kind')
        handler = type_builders.get(type_kind)
        if not handler:
            raise Exception(
                'Invalid or incomplete schema, unknown kind: {}. Ensure that a full introspection query '
                'is used in order to build a client schema.'.format(type_kind)
            )

        return handler(type)

    def build_scalar_def(scalar_introspection):
        return GraphQLScalarType(
            name=scalar_introspection['name'],
            description=scalar_introspection.get('description'),
            serialize=_none,
            parse_value=_false,
            parse_literal=_false
        )

    def build_object_def(object_introspection):
        return GraphQLObjectType(
            name=object_introspection['name'],
            description=object_introspection.get('description'),
            interfaces=[get_interface_type(i) for i in object_introspection.get('interfaces', [])],
            fields=lambda: build_field_def_map(object_introspection)
        )

    def build_interface_def(interface_introspection):
        return GraphQLInterfaceType(
            name=interface_introspection['name'],
            description=interface_introspection.get('description'),
            fields=lambda: build_field_def_map(interface_introspection),
            resolve_type=no_execution
        )

    def build_union_def(union_introspection):
        return GraphQLUnionType(
            name=union_introspection['name'],
            description=union_introspection.get('description'),
            types=[get_object_type(t) for t in union_introspection.get('possibleTypes', [])],
            resolve_type=no_execution
        )

    def build_enum_def(enum_introspection):
        return GraphQLEnumType(
            name=enum_introspection['name'],
            description=enum_introspection.get('description'),
            values=OrderedDict([(value_introspection['name'],
                                 GraphQLEnumValue(description=value_introspection.get('description'),
                                                  deprecation_reason=value_introspection.get('deprecationReason')))
                                for value_introspection in enum_introspection.get('enumValues', [])
                                ])
        )

    def build_input_object_def(input_object_introspection):
        return GraphQLInputObjectType(
            name=input_object_introspection['name'],
            description=input_object_introspection.get('description'),
            fields=lambda: build_input_value_def_map(
                input_object_introspection.get('inputFields'), GraphQLInputObjectField
            )
        )

    type_builders = {
        TypeKind.SCALAR: build_scalar_def,
        TypeKind.OBJECT: build_object_def,
        TypeKind.INTERFACE: build_interface_def,
        TypeKind.UNION: build_union_def,
        TypeKind.ENUM: build_enum_def,
        TypeKind.INPUT_OBJECT: build_input_object_def
    }

    def build_field_def_map(type_introspection):
        return OrderedDict([
            (f['name'], GraphQLField(
                type=get_output_type(f['type']),
                description=f.get('description'),
                resolver=no_execution,
                deprecation_reason=f.get('deprecationReason'),
                args=build_input_value_def_map(f.get('args'), GraphQLArgument)))
            for f in type_introspection.get('fields', [])
        ])

    def build_default_value(f):
        default_value = f.get('defaultValue')
        if default_value is None:
            return None

        return value_from_ast(parse_value(default_value), get_input_type(f['type']))

    def build_input_value_def_map(input_value_introspection, argument_type):
        return OrderedDict([
            (f['name'], build_input_value(f, argument_type)) for f in input_value_introspection
        ])

    def build_input_value(input_value_introspection, argument_type):
        input_value = argument_type(
            description=input_value_introspection['description'],
            type=get_input_type(input_value_introspection['type']),
            default_value=build_default_value(input_value_introspection)
        )
        return input_value

    def build_directive(directive_introspection):
        # Support deprecated `on****` fields for building `locations`, as this
        # is used by GraphiQL which may need to support outdated servers.
        locations = list(directive_introspection.get('locations', []))
        if not locations:
            locations = []
            if directive_introspection.get('onField', False):
                locations += list(DirectiveLocation.FIELD_LOCATIONS)
            if directive_introspection.get('onOperation', False):
                locations += list(DirectiveLocation.OPERATION_LOCATIONS)
            if directive_introspection.get('onFragment', False):
                locations += list(DirectiveLocation.FRAGMENT_LOCATIONS)

        return GraphQLDirective(
            name=directive_introspection['name'],
            description=directive_introspection.get('description'),
            # TODO: {} ?
            args=build_input_value_def_map(directive_introspection.get('args', {}), GraphQLArgument),
            locations=locations
        )

    # Iterate through all types, getting the type definition for each, ensuring
    # that any type not directly referenced by a field will get created.
    types = [get_named_type(type_introspection_name) for type_introspection_name in type_introspection_map.keys()]

    query_type = get_object_type(schema_introspection['queryType'])
    mutation_type = get_object_type(
        schema_introspection['mutationType']) if schema_introspection.get('mutationType') else None
    subscription_type = get_object_type(schema_introspection['subscriptionType']) if \
        schema_introspection.get('subscriptionType') else None

    directives = [build_directive(d) for d in schema_introspection['directives']] \
        if schema_introspection['directives'] else []

    return GraphQLSchema(
        query=query_type,
        mutation=mutation_type,
        subscription=subscription_type,
        directives=directives,
        types=types
    )



import base64
import json
import logging
import time
from typing import Any, Dict, List, Optional, Tuple

import wandb
from wandb.apis.internal import Api
from wandb.sdk.launch.builder.abstract import AbstractBuilder
from wandb.sdk.launch.environment.abstract import AbstractEnvironment
from wandb.sdk.launch.registry.abstract import AbstractRegistry
from wandb.sdk.launch.registry.local_registry import LocalRegistry
from wandb.util import get_module

from .._project_spec import EntryPoint, LaunchProject
from ..builder.build import get_env_vars_dict
from ..utils import (
    LOG_PREFIX,
    PROJECT_SYNCHRONOUS,
    LaunchError,
    get_kube_context_and_api_client,
    make_name_dns_safe,
)
from .abstract import AbstractRun, AbstractRunner, Status

get_module(
    "kubernetes",
    required="Kubernetes runner requires the kubernetes package. Please install it with `pip install wandb[launch]`.",
)

from kubernetes import client  # type: ignore # noqa: E402
from kubernetes.client.api.batch_v1_api import BatchV1Api  # type: ignore # noqa: E402
from kubernetes.client.api.core_v1_api import CoreV1Api  # type: ignore # noqa: E402
from kubernetes.client.models.v1_job import V1Job  # type: ignore # noqa: E402
from kubernetes.client.models.v1_secret import V1Secret  # type: ignore # noqa: E402

TIMEOUT = 5
MAX_KUBERNETES_RETRIES = (
    60  # default 10 second loop time on the agent, this is 10 minutes
)
FAIL_MESSAGE_INTERVAL = 60

_logger = logging.getLogger(__name__)


class KubernetesSubmittedRun(AbstractRun):
    def __init__(
        self,
        batch_api: "BatchV1Api",
        core_api: "CoreV1Api",
        name: str,
        pod_names: List[str],
        namespace: Optional[str] = "default",
        secret: Optional["V1Secret"] = None,
    ) -> None:
        self.batch_api = batch_api
        self.core_api = core_api
        self.name = name
        self.namespace = namespace
        self.job = self.batch_api.read_namespaced_job(
            name=self.name, namespace=self.namespace
        )
        self._fail_count = 0
        self.pod_names = pod_names
        self.secret = secret

    @property
    def id(self) -> str:
        return self.name

    def get_job(self) -> "V1Job":
        return self.batch_api.read_namespaced_job(
            name=self.name, namespace=self.namespace
        )

    def wait(self) -> bool:
        while True:
            status = self.get_status()
            wandb.termlog(f"{LOG_PREFIX}Job {self.name} status: {status}")
            if status.state != "running":
                break
            time.sleep(5)
        return (
            status.state == "finished"
        )  # todo: not sure if this (copied from aws runner) is the right approach? should we return false on failure

    def get_status(self) -> Status:
        job_response = self.batch_api.read_namespaced_job_status(
            name=self.name, namespace=self.namespace
        )
        status = job_response.status

        pod = self.core_api.read_namespaced_pod(
            name=self.pod_names[0], namespace=self.namespace
        )
        if pod.status.phase in ["Pending", "Unknown"]:
            now = time.time()
            if self._fail_count == 0:
                self._fail_first_msg_time = now
                self._fail_last_msg_time = 0.0
            self._fail_count += 1
            if now - self._fail_last_msg_time > FAIL_MESSAGE_INTERVAL:
                wandb.termlog(
                    f"{LOG_PREFIX}Pod has not started yet for job: {self.name}. Will wait up to {round(10 - (now - self._fail_first_msg_time)/60)} minutes."
                )
                self._fail_last_msg_time = now
            if self._fail_count > MAX_KUBERNETES_RETRIES:
                raise LaunchError(f"Failed to start job {self.name}")
        # todo: we only handle the 1 pod case. see https://kubernetes.io/docs/concepts/workloads/controllers/job/#parallel-jobs for multipod handling
        return_status = None
        if status.succeeded == 1:
            return_status = Status("finished")
        elif status.failed is not None and status.failed >= 1:
            return_status = Status("failed")
        elif status.active == 1:
            return Status("running")
        elif status.conditions is not None and status.conditions[0].type == "Suspended":
            return_status = Status("stopped")
        else:
            return_status = Status("unknown")
        if (
            return_status.state in ["stopped", "failed", "finished"]
            and self.secret is not None
        ):
            try:
                self.core_api.delete_namespaced_secret(
                    self.secret.metadata.name, self.namespace
                )
            except Exception as e:
                wandb.termerror(
                    f"Error deleting secret {self.secret.metadata.name}: {str(e)}"
                )
        return return_status

    def suspend(self) -> None:
        self.job.spec.suspend = True
        self.batch_api.patch_namespaced_job(
            name=self.name, namespace=self.namespace, body=self.job
        )
        timeout = TIMEOUT
        job_response = self.batch_api.read_namespaced_job_status(
            name=self.name, namespace=self.namespace
        )
        while job_response.status.conditions is None and timeout > 0:
            time.sleep(1)
            timeout -= 1
            job_response = self.batch_api.read_namespaced_job_status(
                name=self.name, namespace=self.namespace
            )

        if timeout == 0 or job_response.status.conditions[0].type != "Suspended":
            raise LaunchError(
                "Failed to suspend job {}. Check Kubernetes dashboard for more info.".format(
                    self.name
                )
            )

    def cancel(self) -> None:
        self.suspend()
        self.batch_api.delete_namespaced_job(name=self.name, namespace=self.namespace)


class KubernetesRunner(AbstractRunner):
    def __init__(
        self, api: Api, backend_config: Dict[str, Any], environment: AbstractEnvironment
    ) -> None:
        super().__init__(api, backend_config)
        self.environment = environment

    def wait_job_launch(
        self, job_name: str, namespace: str, core_api: "CoreV1Api"
    ) -> List[str]:
        pods = core_api.list_namespaced_pod(
            label_selector=f"job-name={job_name}", namespace=namespace
        )
        timeout = TIMEOUT
        while len(pods.items) == 0 and timeout > 0:
            time.sleep(1)
            timeout -= 1
            pods = core_api.list_namespaced_pod(
                label_selector=f"job-name={job_name}", namespace=namespace
            )

        if timeout == 0:
            raise LaunchError(
                "No pods found for job {}. Check dashboard to see if job was launched successfully.".format(
                    job_name
                )
            )

        pod_names = [pi.metadata.name for pi in pods.items]
        wandb.termlog(
            f"{LOG_PREFIX}Job {job_name} created on pod(s) {', '.join(pod_names)}. See logs with e.g. `kubectl logs {pod_names[0]} -n {namespace}`."
        )
        return pod_names

    def get_namespace(
        self, resource_args: Dict[str, Any], context: Dict[str, Any]
    ) -> str:
        default_namespace = (
            context["context"].get("namespace", "default") if context else "default"
        )
        return (  # type: ignore[no-any-return]
            self.backend_config.get("runner", {}).get("namespace")
            or resource_args.get("namespace")
            or default_namespace
        )

    def _inject_defaults(
        self,
        resource_args: Dict[str, Any],
        launch_project: LaunchProject,
        builder: Optional[AbstractBuilder],
        namespace: str,
        core_api: "CoreV1Api",
    ) -> Tuple[Dict[str, Any], Optional["V1Secret"]]:
        """Apply our default values, return job dict and secret."""
        job: Dict[str, Any] = {
            "apiVersion": "batch/v1",
            "kind": "Job",
        }
        job.update(resource_args)

        job_metadata: Dict[str, Any] = job.get("metadata", {})
        job_spec: Dict[str, Any] = {"backoffLimit": 0, "ttlSecondsAfterFinished": 60}
        job_spec.update(job.get("spec", {}))
        pod_template: Dict[str, Any] = job_spec.get("template", {})
        pod_spec: Dict[str, Any] = {"restartPolicy": "Never"}
        pod_spec.update(pod_template.get("spec", {}))
        containers: List[Dict[str, Any]] = pod_spec.get("containers", [{}])

        # name precedence: name in spec > generated name
        if not job_metadata.get("name"):
            job_metadata["generateName"] = make_name_dns_safe(
                f"launch-{launch_project.target_entity}-{launch_project.target_project}-"
            )

        for i, cont in enumerate(containers):
            if "name" not in cont:
                cont["name"] = cont.get("name", "launch" + str(i))
            if "securityContext" not in cont:
                cont["securityContext"] = {
                    "allowPrivilegeEscalation": False,
                    "capabilities": {"drop": ["ALL"]},
                    "seccompProfile": {"type": "RuntimeDefault"},
                }

        secret = None
        entry_point = launch_project.get_single_entry_point()
        if launch_project.docker_image:
            if len(containers) > 1:
                raise LaunchError(
                    "Invalid specification of multiple containers. See https://docs.wandb.ai/guides/launch for guidance on submitting jobs."
                )
            # dont specify run id if user provided image, could have multiple runs
            containers[0]["image"] = launch_project.docker_image
            # TODO: handle secret pulling image from registry
        elif not any(["image" in cont for cont in containers]):
            if len(containers) > 1:
                raise LaunchError(
                    "Launch only builds one container at a time. See https://docs.wandb.ai/guides/launch for guidance on submitting jobs."
                )
            assert entry_point is not None
            assert builder is not None
            image_uri = builder.build_image(launch_project, entry_point)
            # in the non instance case we need to make an imagePullSecret
            # so the new job can pull the image
            if not builder.registry:
                raise LaunchError(
                    "No registry specified. Please specify a registry in your wandb/settings file or pass a registry to the builder."
                )
            secret = maybe_create_imagepull_secret(
                core_api, builder.registry, launch_project.run_id, namespace
            )
            if secret is not None:
                pod_spec["imagePullSecrets"] = [
                    {"name": f"regcred-{launch_project.run_id}"}
                ]

            containers[0]["image"] = image_uri

        inject_entrypoint_and_args(
            containers,
            entry_point,
            launch_project.override_args,
            launch_project.override_entrypoint is not None,
        )

        env_vars = get_env_vars_dict(launch_project, self._api)
        for cont in containers:
            # Add our env vars to user supplied env vars
            env = cont.get("env", [])
            env.extend(
                [{"name": key, "value": value} for key, value in env_vars.items()]
            )
            cont["env"] = env

        pod_spec["containers"] = containers
        pod_template["spec"] = pod_spec
        job_spec["template"] = pod_template
        job["spec"] = job_spec
        job["metadata"] = job_metadata

        return job, secret

    def run(
        self,
        launch_project: LaunchProject,
        builder: Optional[AbstractBuilder],
    ) -> Optional[AbstractRun]:  # noqa: C901
        kubernetes = get_module(  # noqa: F811
            "kubernetes",
            required="Kubernetes runner requires the kubernetes package. Please"
            " install it with `pip install wandb[launch]`.",
        )
        resource_args = launch_project.resource_args.get("kubernetes", {})
        if not resource_args:
            wandb.termlog(
                f"{LOG_PREFIX}Note: no resource args specified. Add a Kubernetes yaml spec or other options in a json file with --resource-args <json>."
            )
        _logger.info(f"Running Kubernetes job with resource args: {resource_args}")

        context, api_client = get_kube_context_and_api_client(kubernetes, resource_args)

        batch_api = kubernetes.client.BatchV1Api(api_client)
        core_api = kubernetes.client.CoreV1Api(api_client)

        namespace = self.get_namespace(resource_args, context)

        job, secret = self._inject_defaults(
            resource_args,
            launch_project,
            builder,
            namespace,
            core_api,
        )

        msg = "Creating Kubernetes job"
        if "name" in resource_args:
            msg += f": {resource_args['name']}"
        _logger.info(msg)
        job_response = kubernetes.utils.create_from_yaml(
            api_client, yaml_objects=[job], namespace=namespace
        )[0][
            0
        ]  # create_from_yaml returns a nested list of k8s objects
        job_name = job_response.metadata.name

        pod_names = self.wait_job_launch(job_name, namespace, core_api)

        submitted_job = KubernetesSubmittedRun(
            batch_api, core_api, job_name, pod_names, namespace, secret
        )

        if self.backend_config[PROJECT_SYNCHRONOUS]:
            submitted_job.wait()

        return submitted_job


def inject_entrypoint_and_args(
    containers: List[dict],
    entry_point: Optional[EntryPoint],
    override_args: List[str],
    should_override_entrypoint: bool,
) -> None:
    for i in range(len(containers)):
        if override_args:
            containers[i]["args"] = override_args
        if entry_point and (
            not containers[i].get("command") or should_override_entrypoint
        ):
            containers[i]["command"] = entry_point.command


def maybe_create_imagepull_secret(
    core_api: "CoreV1Api",
    registry: AbstractRegistry,
    run_id: str,
    namespace: str,
) -> Optional["V1Secret"]:
    secret = None
    if isinstance(registry, LocalRegistry):
        # Secret not required
        return None
    uname, token = registry.get_username_password()
    creds_info = {
        "auths": {
            registry.uri: {
                "auth": base64.b64encode(f"{uname}:{token}".encode()).decode(),
                # need an email but the use is deprecated
                "email": "deprecated@wandblaunch.com",
            }
        }
    }
    secret_data = {
        ".dockerconfigjson": base64.b64encode(json.dumps(creds_info).encode()).decode()
    }
    secret = client.V1Secret(
        data=secret_data,
        metadata=client.V1ObjectMeta(name=f"regcred-{run_id}", namespace=namespace),
        kind="Secret",
        type="kubernetes.io/dockerconfigjson",
    )
    try:
        return core_api.create_namespaced_secret(namespace, secret)
    except Exception as e:
        raise LaunchError(f"Exception when creating Kubernetes secret: {str(e)}\n")



#
import os
from typing import Callable, Generator

WANDB_DIRS = ("wandb", ".wandb")

CONFIG_FNAME = "config.yaml"
OUTPUT_FNAME = "output.log"
DIFF_FNAME = "diff.patch"
SUMMARY_FNAME = "wandb-summary.json"
METADATA_FNAME = "wandb-metadata.json"
REQUIREMENTS_FNAME = "requirements.txt"
HISTORY_FNAME = "wandb-history.jsonl"
EVENTS_FNAME = "wandb-events.jsonl"
JOBSPEC_FNAME = "wandb-jobspec.json"
CONDA_ENVIRONMENTS_FNAME = "conda-environment.yaml"


def is_wandb_file(name: str) -> bool:
    return (
        name.startswith("wandb")
        or name == METADATA_FNAME
        or name == CONFIG_FNAME
        or name == REQUIREMENTS_FNAME
        or name == OUTPUT_FNAME
        or name == DIFF_FNAME
        or name == CONDA_ENVIRONMENTS_FNAME
    )


def filtered_dir(
    root: str, include_fn: Callable[[str], bool], exclude_fn: Callable[[str], bool]
) -> Generator[str, None, None]:
    """Simple generator to walk a directory."""
    for dirpath, _, files in os.walk(root):
        for fname in files:
            file_path = os.path.join(dirpath, fname)
            if include_fn(file_path) and not exclude_fn(file_path):
                yield file_path


def exclude_wandb_fn(path: str) -> bool:
    return any(os.sep + wandb_dir + os.sep in path for wandb_dir in WANDB_DIRS)



import pytest
import requests
import vcr

from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport

# https://github.com/graphql-python/swapi-graphene
URL = 'http://127.0.0.1:8000/graphql'


@pytest.fixture
def client():
    with vcr.use_cassette('tests/fixtures/vcr_cassettes/client.yaml'):
        request = requests.get(
            URL,
            headers={
                'Host': 'swapi.graphene-python.org',
                'Accept': 'text/html',
            }
        )
        request.raise_for_status()
        csrf = request.cookies['csrftoken']

        return Client(
            transport=RequestsHTTPTransport(
                url=URL,
                cookies={"csrftoken": csrf},
                headers={'x-csrftoken':  csrf}),
            fetch_schema_from_transport=True
        )


def test_hero_name_query(client):
    query = gql('''
    {
      myFavoriteFilm: film(id:"RmlsbToz") {
        id
        title
        episodeId
        characters(first:5) {
          edges {
            node {
              name
            }
          }
        }
      }
    }
    ''')
    expected = {
        "myFavoriteFilm": {
            "id": "RmlsbToz",
            "title": "Return of the Jedi",
            "episodeId": 6,
            "characters": {
                "edges": [
                  {
                      "node": {
                          "name": "Luke Skywalker"
                      }
                  },
                    {
                      "node": {
                          "name": "C-3PO"
                      }
                  },
                    {
                      "node": {
                          "name": "R2-D2"
                      }
                  },
                    {
                      "node": {
                          "name": "Darth Vader"
                      }
                  },
                    {
                      "node": {
                          "name": "Leia Organa"
                      }
                  }
                ]
            }
        }
    }
    with vcr.use_cassette('tests/fixtures/vcr_cassettes/execute.yaml'):
        result = client.execute(query)
        assert result == expected



__all__ = ("WandbMetricsLogger", "WandbModelCheckpoint", "WandbEvalCallback")

from .metrics_logger import WandbMetricsLogger
from .model_checkpoint import WandbModelCheckpoint
from .tables_builder import WandbEvalCallback



from typing import TYPE_CHECKING, Dict, Optional, Sequence, Type, Union

from wandb.sdk.lib.paths import FilePathStr, URIStr

if TYPE_CHECKING:
    from urllib.parse import ParseResult

    from wandb.filesync.step_prepare import StepPrepare
    from wandb.sdk.interface.artifacts import Artifact, ArtifactManifestEntry
    from wandb.sdk.internal.progress import ProgressFn


class StorageLayout:
    V1 = "V1"
    V2 = "V2"


class StoragePolicy:
    @classmethod
    def lookup_by_name(cls, name: str) -> Optional[Type["StoragePolicy"]]:
        for sub in cls.__subclasses__():
            if sub.name() == name:
                return sub
        return None

    @classmethod
    def name(cls) -> str:
        raise NotImplementedError

    @classmethod
    def from_config(cls, config: Dict) -> "StoragePolicy":
        raise NotImplementedError

    def config(self) -> Dict:
        raise NotImplementedError

    def load_file(
        self, artifact: "Artifact", manifest_entry: "ArtifactManifestEntry"
    ) -> str:
        raise NotImplementedError

    def store_file_sync(
        self,
        artifact_id: str,
        artifact_manifest_id: str,
        entry: "ArtifactManifestEntry",
        preparer: "StepPrepare",
        progress_callback: Optional["ProgressFn"] = None,
    ) -> bool:
        raise NotImplementedError

    async def store_file_async(
        self,
        artifact_id: str,
        artifact_manifest_id: str,
        entry: "ArtifactManifestEntry",
        preparer: "StepPrepare",
        progress_callback: Optional["ProgressFn"] = None,
    ) -> bool:
        """Async equivalent to `store_file_sync`."""
        raise NotImplementedError

    def store_reference(
        self,
        artifact: "Artifact",
        path: Union[URIStr, FilePathStr],
        name: Optional[str] = None,
        checksum: bool = True,
        max_objects: Optional[int] = None,
    ) -> Sequence["ArtifactManifestEntry"]:
        raise NotImplementedError

    def load_reference(
        self,
        manifest_entry: "ArtifactManifestEntry",
        local: bool = False,
    ) -> str:
        raise NotImplementedError


class StorageHandler:
    def can_handle(self, parsed_url: "ParseResult") -> bool:
        """Checks whether this handler can handle the given url.

        Returns:
            Whether this handler can handle the given url.
        """
        raise NotImplementedError

    def load_path(
        self,
        manifest_entry: "ArtifactManifestEntry",
        local: bool = False,
    ) -> Union[URIStr, FilePathStr]:
        """Load a file or directory given the corresponding index entry.

        Args:
            manifest_entry: The index entry to load
            local: Whether to load the file locally or not

        Returns:
            A path to the file represented by `index_entry`
        """
        raise NotImplementedError

    def store_path(
        self,
        artifact: "Artifact",
        path: Union[URIStr, FilePathStr],
        name: Optional[str] = None,
        checksum: bool = True,
        max_objects: Optional[int] = None,
    ) -> Sequence["ArtifactManifestEntry"]:
        """Store the file or directory at the given path to the specified artifact.

        Args:
            path: The path to store
            name: If specified, the logical name that should map to `path`
            checksum: Whether to compute the checksum of the file
            max_objects: The maximum number of objects to store

        Returns:
            A list of manifest entries to store within the artifact
        """
        raise NotImplementedError



"""static settings."""
from typing import Any, Dict, Iterable, Mapping, Optional, Sequence, Tuple, Union

SettingsDict = Dict[str, Union[str, float, Tuple, None]]


class SettingsStatic:
    # TODO(jhr): figure out how to share type defs with sdk/wandb_settings.py
    _offline: Optional[bool]
    _sync: bool
    _disable_stats: Optional[bool]
    _disable_meta: Optional[bool]
    _flow_control: bool
    _start_time: float
    _start_datetime: str
    _stats_pid: int
    _stats_sample_rate_seconds: float
    _stats_samples_to_average: int
    _stats_join_assets: bool
    _stats_neuron_monitor_config_path: Optional[str]
    _stats_open_metrics_endpoints: Mapping[str, str]
    _stats_open_metrics_filters: Union[
        Sequence[str], Mapping[str, Mapping[str, str]], None
    ]
    files_dir: str
    program_relpath: Optional[str]
    log_internal: str
    _internal_check_process: bool
    is_local: Optional[bool]
    _colab: Optional[bool]
    _jupyter: Optional[bool]
    notebook_name: Optional[str]
    _jupyter_path: Optional[str]
    _jupyter_name: Optional[str]
    _jupyter_root: Optional[str]
    _network_buffer: Optional[int]
    _disable_service: Optional[bool]
    _live_policy_rate_limit: Optional[int]
    resume: Optional[str]
    program: Optional[str]
    silent: Optional[bool]
    email: Optional[str]
    git_commit: Optional[str]
    git_remote: Optional[str]
    git_remote_url: Optional[str]
    git_root: Optional[str]
    docker: Optional[str]
    _cuda: Optional[str]
    _args: Optional[Sequence[str]]
    _os: Optional[str]
    _python: Optional[str]
    disable_git: Optional[bool]
    _save_requirements: Optional[bool]
    save_code: Optional[bool]
    disable_code: Optional[bool]
    anonymous: Optional[str]
    host: Optional[str]
    username: Optional[str]
    _executable: str
    run_url: Optional[str]
    run_name: Optional[str]
    sync_file: str
    _flow_control_disabled: bool
    _flow_control_custom: bool
    disable_job_creation: bool
    _async_upload_concurrency_limit: Optional[int]
    _extra_http_headers: Optional[Mapping[str, str]]

    # TODO(jhr): clean this up, it is only in SettingsStatic and not in Settings
    _log_level: int

    def __init__(self, d: "SettingsDict") -> None:
        object.__setattr__(self, "__dict__", d)

    def __setattr__(self, name: str, value: object) -> None:
        raise AttributeError("Error: SettingsStatic is a readonly object")

    def __setitem__(self, key: str, val: object) -> None:
        raise AttributeError("Error: SettingsStatic is a readonly object")

    def keys(self) -> "Iterable[str]":
        return self.__dict__.keys()

    def items(self) -> "Iterable[Any]":
        return self.__dict__.items()

    def __getitem__(self, key: str) -> "Any":
        return self.__dict__[key]

    def __str__(self) -> str:
        return str(self.__dict__)

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        return self.__dict__.get(key, default)



# from promise import Promise
# from promise.dataloader import DataLoader


# def id_loader(**options):
#     load_calls = []

#     def fn(keys):
#         load_calls.append(keys)
#         return Promise.resolve(keys)

#     identity_loader = DataLoader(fn, **options)
#     return identity_loader, load_calls


# def test_batches_multiple_requests():
#     identity_loader, load_calls = id_loader()

#     @Promise.safe
#     def safe():
#         promise1 = identity_loader.load(1)
#         promise2 = identity_loader.load(2)
#         return promise1, promise2

#     promise1, promise2 = safe()
#     value1, value2 = Promise.all([promise1, promise2]).get()
#     assert value1 == 1
#     assert value2 == 2

#     assert load_calls == [[1, 2]]


# def test_batches_multiple_requests_two():
#     identity_loader, load_calls = id_loader()

#     @Promise.safe
#     def safe():
#         promise1 = identity_loader.load(1)
#         promise2 = identity_loader.load(2)
#         return Promise.all([promise1, promise2])

#     p = safe()
#     value1, value2 = p.get()

#     assert value1 == 1
#     assert value2 == 2

#     assert load_calls == [[1, 2]]


# @Promise.safe
# def test_batches_multiple_requests_safe():
#     identity_loader, load_calls = id_loader()

#     promise1 = identity_loader.load(1)
#     promise2 = identity_loader.load(2)

#     p = Promise.all([promise1, promise2])

#     value1, value2 = p.get()

#     assert value1 == 1
#     assert value2 == 2

#     assert load_calls == [[1, 2]]



#
"""Setup wandb session.

This module configures a wandb session which can extend to mutiple wandb runs.

Functions:
    setup(): Configure wandb session.

Early logging keeps track of logger output until the call to wandb.init() when the
run_id can be resolved.

"""

import logging
import os
import sys
import threading
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import wandb

from . import wandb_manager, wandb_settings
from .lib import config_util, server, tracelog

Settings = Union["wandb.sdk.wandb_settings.Settings", Dict[str, Any]]

Logger = Union[logging.Logger, "_EarlyLogger"]

if TYPE_CHECKING:
    from . import wandb_run

# logger will be configured to be either a standard logger instance or _EarlyLogger
logger: Optional[Logger] = None


def _set_logger(log_object: Logger) -> None:
    """Configure module logger."""
    global logger
    logger = log_object


class _EarlyLogger:
    """Early logger which captures logs in memory until logging can be configured."""

    def __init__(self) -> None:
        self._log: List[tuple] = []
        self._exception: List[tuple] = []
        # support old warn() as alias of warning()
        self.warn = self.warning

    def debug(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.DEBUG, msg, args, kwargs))

    def info(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.INFO, msg, args, kwargs))

    def warning(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.WARNING, msg, args, kwargs))

    def error(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.ERROR, msg, args, kwargs))

    def critical(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((logging.CRITICAL, msg, args, kwargs))

    def exception(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self._exception.append((msg, args, kwargs))

    def log(self, level: str, msg: str, *args: Any, **kwargs: Any) -> None:
        self._log.append((level, msg, args, kwargs))

    def _flush(self) -> None:
        assert self is not logger
        assert logger is not None
        for level, msg, args, kwargs in self._log:
            logger.log(level, msg, *args, **kwargs)
        for msg, args, kwargs in self._exception:
            logger.exception(msg, *args, **kwargs)


class _WandbSetup__WandbSetup:  # noqa: N801
    """Inner class of _WandbSetup."""

    _manager: Optional[wandb_manager._Manager]
    _pid: int

    def __init__(
        self,
        pid: int,
        settings: Optional[Settings] = None,
        environ: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._environ = environ or dict(os.environ)
        self._sweep_config: Optional[Dict[str, Any]] = None
        self._config: Optional[Dict[str, Any]] = None
        self._server: Optional[server.Server] = None
        self._manager: Optional[wandb_manager._Manager] = None
        self._pid = pid

        # keep track of multiple runs, so we can unwind with join()s
        self._global_run_stack: List["wandb_run.Run"] = []

        # TODO(jhr): defer strict checks until settings are fully initialized
        #            and logging is ready
        self._early_logger = _EarlyLogger()
        _set_logger(self._early_logger)

        self._settings = self._settings_setup(settings, self._early_logger)
        # self._settings.freeze()

        wandb.termsetup(self._settings, logger)

        self._check()
        self._setup()

        tracelog_mode = self._settings._tracelog
        if tracelog_mode:
            tracelog.enable(tracelog_mode)

    def _settings_setup(
        self,
        settings: Optional[Settings] = None,
        early_logger: Optional[_EarlyLogger] = None,
    ) -> "wandb_settings.Settings":
        s = wandb_settings.Settings()
        s._apply_base(pid=self._pid, _logger=early_logger)
        s._apply_config_files(_logger=early_logger)
        s._apply_env_vars(self._environ, _logger=early_logger)

        if isinstance(settings, wandb_settings.Settings):
            s._apply_settings(settings, _logger=early_logger)
        elif isinstance(settings, dict):
            # if passed settings arg is a mapping, update the settings with it
            s._apply_setup(settings, _logger=early_logger)

        s._infer_settings_from_environment()
        if not s._cli_only_mode:
            s._infer_run_settings_from_environment(_logger=early_logger)

        return s

    def _update(
        self,
        settings: Optional[Settings] = None,
    ) -> None:
        if settings is None:
            return
        # self._settings.unfreeze()
        if isinstance(settings, wandb_settings.Settings):
            # todo: check the logic here. this _only_ comes up in tests?
            self._settings._apply_settings(settings)
        elif isinstance(settings, dict):
            # if it is a mapping, update the settings with it
            self._settings.update(settings, source=wandb_settings.Source.SETUP)
        # self._settings.freeze()

    def _update_user_settings(self, settings: Optional[Settings] = None) -> None:
        settings = settings or self._settings
        # Get rid of cached results to force a refresh.
        self._server = None
        user_settings = self._load_user_settings(settings=settings)
        if user_settings is not None:
            # self._settings.unfreeze()
            self._settings._apply_user(user_settings)
            # self._settings.freeze()

    def _early_logger_flush(self, new_logger: Logger) -> None:
        if not self._early_logger:
            return
        _set_logger(new_logger)
        # self._settings._clear_early_logger()
        self._early_logger._flush()

    def _get_logger(self) -> Optional[Logger]:
        return logger

    @property
    def settings(self) -> "wandb_settings.Settings":
        return self._settings

    def _get_entity(self) -> Optional[str]:
        if self._settings and self._settings._offline:
            return None
        if self._server is None:
            self._load_viewer()
        assert self._server is not None
        entity = self._server._viewer.get("entity")
        return entity

    def _get_username(self) -> Optional[str]:
        if self._settings and self._settings._offline:
            return None
        if self._server is None:
            self._load_viewer()
        assert self._server is not None
        username = self._server._viewer.get("username")
        return username

    def _get_teams(self) -> List[str]:
        if self._settings and self._settings._offline:
            return []
        if self._server is None:
            self._load_viewer()
        assert self._server is not None
        teams = self._server._viewer.get("teams")
        if teams:
            teams = [team["node"]["name"] for team in teams["edges"]]
        return teams or []

    def _load_viewer(self, settings: Optional[Settings] = None) -> None:
        if self._settings and self._settings._offline:
            return
        if isinstance(settings, dict):
            settings = wandb_settings.Settings(**settings)
        s = server.Server(settings=settings)
        s.query_with_timeout()
        self._server = s

    def _load_user_settings(
        self, settings: Optional[Settings] = None
    ) -> Optional[Dict[str, Any]]:
        if self._server is None:
            self._load_viewer(settings=settings)

        # offline?
        if self._server is None:
            return None

        flags = self._server._flags
        user_settings = dict()
        if "code_saving_enabled" in flags:
            user_settings["save_code"] = flags["code_saving_enabled"]

        email = self._server._viewer.get("email", None)
        if email:
            user_settings["email"] = email

        return user_settings

    def _check(self) -> None:
        if hasattr(threading, "main_thread"):
            if threading.current_thread() is not threading.main_thread():
                pass
        elif threading.current_thread().name != "MainThread":
            print("bad thread2", threading.current_thread().name)
        if getattr(sys, "frozen", False):
            print("frozen, could be trouble")

    def _setup(self) -> None:
        self._setup_manager()

        sweep_path = self._settings.sweep_param_path
        if sweep_path:
            self._sweep_config = config_util.dict_from_config_file(
                sweep_path, must_exist=True
            )

        # if config_paths was set, read in config dict
        if self._settings.config_paths:
            # TODO(jhr): handle load errors, handle list of files
            for config_path in self._settings.config_paths:
                config_dict = config_util.dict_from_config_file(config_path)
                if config_dict is None:
                    continue
                if self._config is not None:
                    self._config.update(config_dict)
                else:
                    self._config = config_dict

    def _teardown(self, exit_code: Optional[int] = None) -> None:
        exit_code = exit_code or 0
        self._teardown_manager(exit_code=exit_code)

    def _setup_manager(self) -> None:
        if self._settings._disable_service:
            return
        self._manager = wandb_manager._Manager(settings=self._settings)

    def _teardown_manager(self, exit_code: int) -> None:
        if not self._manager:
            return
        self._manager._teardown(exit_code)
        self._manager = None

    def _get_manager(self) -> Optional[wandb_manager._Manager]:
        return self._manager


class _WandbSetup:
    """Wandb singleton class.

    Note: This is a process local singleton.
    (Forked processes will get a new copy of the object)
    """

    _instance: Optional["_WandbSetup__WandbSetup"] = None

    def __init__(self, settings: Optional[Settings] = None) -> None:
        pid = os.getpid()
        if _WandbSetup._instance and _WandbSetup._instance._pid == pid:
            _WandbSetup._instance._update(settings=settings)
            return
        _WandbSetup._instance = _WandbSetup__WandbSetup(settings=settings, pid=pid)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._instance, name)


def _setup(
    settings: Optional[Settings] = None,
    _reset: bool = False,
) -> Optional["_WandbSetup"]:
    """Set up library context."""
    if _reset:
        setup_instance = _WandbSetup._instance
        if setup_instance:
            setup_instance._teardown()
        _WandbSetup._instance = None
        return None
    wl = _WandbSetup(settings=settings)
    return wl


def setup(
    settings: Optional[Settings] = None,
) -> Optional["_WandbSetup"]:
    ret = _setup(settings=settings)
    return ret


def teardown(exit_code: Optional[int] = None) -> None:
    setup_instance = _WandbSetup._instance
    if setup_instance:
        setup_instance._teardown(exit_code=exit_code)
    _WandbSetup._instance = None



"""Implementation of the docker builder."""
import logging
import os
from typing import Any, Dict

import wandb
import wandb.docker as docker
from wandb.sdk.launch.builder.abstract import AbstractBuilder
from wandb.sdk.launch.environment.abstract import AbstractEnvironment
from wandb.sdk.launch.registry.abstract import AbstractRegistry

from .._project_spec import (
    EntryPoint,
    LaunchProject,
    create_metadata_file,
    get_entry_point_command,
)
from ..registry.local_registry import LocalRegistry
from ..utils import (
    LOG_PREFIX,
    LaunchDockerError,
    LaunchError,
    sanitize_wandb_api_key,
    warn_failed_packages_from_build_logs,
)
from .build import (
    _create_docker_build_ctx,
    generate_dockerfile,
    image_tag_from_dockerfile_and_source,
    validate_docker_installation,
)

_GENERATED_DOCKERFILE_NAME = "Dockerfile.wandb-autogenerated"
_logger = logging.getLogger(__name__)


class DockerBuilder(AbstractBuilder):
    """Builds a docker image for a project.

    Attributes:
        builder_config (Dict[str, Any]): The builder config.

    """

    builder_type = "docker"
    base_image = "python:3.8"
    target_platform = "linux/amd64"

    def __init__(
        self,
        environment: AbstractEnvironment,
        registry: AbstractRegistry,
        config: Dict[str, Any],
        verify: bool = True,
        login: bool = True,
    ):
        """Initialize a DockerBuilder.

        Arguments:
            environment (AbstractEnvironment): The environment to use.
            registry (AbstractRegistry): The registry to use.
            verify (bool, optional): Whether to verify the functionality of the builder.
            login (bool, optional): Whether to login to the registry.

        Raises:
            LaunchError: If docker is not installed
        """
        self.environment = environment  # Docker builder doesn't actually use this.
        self.registry = registry
        self.config = config
        if verify:
            self.verify()
        if login:
            self.login()

    @classmethod
    def from_config(
        cls,
        config: Dict[str, Any],
        environment: AbstractEnvironment,
        registry: AbstractRegistry,
        verify: bool = True,
    ) -> "DockerBuilder":
        """Create a DockerBuilder from a config.

        Arguments:
            config (Dict[str, Any]): The config.
            registry (AbstractRegistry): The registry to use.
            verify (bool, optional): Whether to verify the functionality of the builder.
            login (bool, optional): Whether to login to the registry.

        Returns:
            DockerBuilder: The DockerBuilder.
        """
        # TODO the config for the docker builder as of yet is empty
        # but ultimately we should add things like target platform, base image, etc.
        return cls(environment, registry, config)

    def verify(self) -> None:
        """Verify the builder."""
        validate_docker_installation()

    def login(self) -> None:
        """Login to the registry."""
        if isinstance(self.registry, LocalRegistry):
            _logger.info(f"{LOG_PREFIX}No registry configured, skipping login.")
        else:
            username, password = self.registry.get_username_password()
            docker.login(username, password, self.registry.uri)

    def build_image(
        self,
        launch_project: LaunchProject,
        entrypoint: EntryPoint,
    ) -> str:
        """Build the image for the given project.

        Arguments:
            launch_project (LaunchProject): The project to build.
            entrypoint (EntryPoint): The entrypoint to use.
        """
        dockerfile_str = generate_dockerfile(
            launch_project, entrypoint, launch_project.resource, "docker"
        )

        image_tag = image_tag_from_dockerfile_and_source(launch_project, dockerfile_str)

        repository = None if not self.registry else self.registry.get_repo_uri()
        # if repo is set, use the repo name as the image name
        if repository:
            image_uri = f"{repository}:{image_tag}"
        # otherwise, base the image name off of the source
        # which the launch_project checks in image_name
        else:
            image_uri = f"{launch_project.image_name}:{image_tag}"

        if not launch_project.build_required() and self.registry.check_image_exists(
            image_uri
        ):
            return image_uri

        _logger.info(
            f"image {image_uri} does not already exist in repository, building."
        )

        entry_cmd = get_entry_point_command(entrypoint, launch_project.override_args)

        create_metadata_file(
            launch_project,
            image_uri,
            sanitize_wandb_api_key(" ".join(entry_cmd)),
            dockerfile_str,
        )
        build_ctx_path = _create_docker_build_ctx(launch_project, dockerfile_str)
        dockerfile = os.path.join(build_ctx_path, _GENERATED_DOCKERFILE_NAME)
        try:
            output = docker.build(
                tags=[image_uri],
                file=dockerfile,
                context_path=build_ctx_path,
                platform=self.config.get("platform"),
            )
            warn_failed_packages_from_build_logs(output, image_uri)

        except docker.DockerError as e:
            raise LaunchDockerError(f"Error communicating with docker client: {e}")

        try:
            os.remove(build_ctx_path)
        except Exception:
            _msg = f"{LOG_PREFIX}Temporary docker context file {build_ctx_path} was not deleted."
            _logger.info(_msg)

        if repository:
            reg, tag = image_uri.split(":")
            wandb.termlog(f"{LOG_PREFIX}Pushing image {image_uri}")
            push_resp = docker.push(reg, tag)
            if push_resp is None:
                raise LaunchError("Failed to push image to repository")
            elif (
                launch_project.resource == "sagemaker"
                and f"The push refers to repository [{repository}]" not in push_resp
            ):
                raise LaunchError(f"Unable to push image to ECR, response: {push_resp}")

        return image_uri



"""catboost init."""

from pathlib import Path
from types import SimpleNamespace
from typing import List, Union

from catboost import CatBoostClassifier, CatBoostRegressor  # type: ignore

import wandb
from wandb.sdk.lib import telemetry as wb_telemetry


class WandbCallback:
    """`WandbCallback` automatically integrates CatBoost with wandb.

    Arguments:
        - metric_period: (int) if you are passing `metric_period` to your CatBoost model please pass the same value here (default=1).

    Passing `WandbCallback` to CatBoost will:
    - log training and validation metrics at every `metric_period`
    - log iteration at every `metric_period`

    Example:
        ```
        train_pool = Pool(train[features], label=train["label"], cat_features=cat_features)
        test_pool = Pool(test[features], label=test["label"], cat_features=cat_features)

        model = CatBoostRegressor(
            iterations=100,
            loss_function="Cox",
            eval_metric="Cox",
        )

        model.fit(
            train_pool,
            eval_set=test_pool,
            callbacks=[WandbCallback()],
        )
        ```
    """

    def __init__(self, metric_period: int = 1):
        if wandb.run is None:
            raise wandb.Error("You must call `wandb.init()` before `WandbCallback()`")

        with wb_telemetry.context() as tel:
            tel.feature.catboost_wandb_callback = True

        self.metric_period: int = metric_period

    def after_iteration(self, info: SimpleNamespace) -> bool:
        if info.iteration % self.metric_period == 0:
            for data, metric in info.metrics.items():
                for metric_name, log in metric.items():
                    # todo: replace with wandb.run._log once available
                    wandb.log({f"{data}-{metric_name}": log[-1]}, commit=False)
            # todo: replace with wandb.run._log once available
            wandb.log({f"iteration@metric-period-{self.metric_period}": info.iteration})

        return True


def _checkpoint_artifact(
    model: Union[CatBoostClassifier, CatBoostRegressor], aliases: List[str]
) -> None:
    """Upload model checkpoint as W&B artifact."""
    if wandb.run is None:
        raise wandb.Error(
            "You must call `wandb.init()` before `_checkpoint_artifact()`"
        )

    model_name = f"model_{wandb.run.id}"
    # save the model in the default `cbm` format
    model_path = Path(wandb.run.dir) / "model"

    model.save_model(model_path)

    model_artifact = wandb.Artifact(name=model_name, type="model")
    model_artifact.add_file(str(model_path))
    wandb.log_artifact(model_artifact, aliases=aliases)


def _log_feature_importance(
    model: Union[CatBoostClassifier, CatBoostRegressor]
) -> None:
    """Log feature importance with default settings."""
    if wandb.run is None:
        raise wandb.Error(
            "You must call `wandb.init()` before `_checkpoint_artifact()`"
        )

    feat_df = model.get_feature_importance(prettified=True)

    fi_data = [
        [feat, feat_imp]
        for feat, feat_imp in zip(feat_df["Feature Id"], feat_df["Importances"])
    ]
    table = wandb.Table(data=fi_data, columns=["Feature", "Importance"])
    # todo: replace with wandb.run._log once available
    wandb.log(
        {
            "Feature Importance": wandb.plot.bar(
                table, "Feature", "Importance", title="Feature Importance"
            )
        },
        commit=False,
    )


def log_summary(
    model: Union[CatBoostClassifier, CatBoostRegressor],
    log_all_params: bool = True,
    save_model_checkpoint: bool = False,
    log_feature_importance: bool = True,
) -> None:
    """`log_summary` logs useful metrics about catboost model after training is done.

    Arguments:
        model: it can be CatBoostClassifier or CatBoostRegressor.
        log_all_params: (boolean) if True (default) log the model hyperparameters as W&B config.
        save_model_checkpoint: (boolean) if True saves the model upload as W&B artifacts.
        log_feature_importance: (boolean) if True (default) logs feature importance as W&B bar chart using the default setting of `get_feature_importance`.

    Using this along with `wandb_callback` will:

    - save the hyperparameters as W&B config,
    - log `best_iteration` and `best_score` as `wandb.summary`,
    - save and upload your trained model to Weights & Biases Artifacts (when `save_model_checkpoint = True`)
    - log feature importance plot.

    Example:
        ```python
        train_pool = Pool(train[features], label=train["label"], cat_features=cat_features)
        test_pool = Pool(test[features], label=test["label"], cat_features=cat_features)

        model = CatBoostRegressor(
            iterations=100,
            loss_function="Cox",
            eval_metric="Cox",
        )

        model.fit(
            train_pool,
            eval_set=test_pool,
            callbacks=[WandbCallback()],
        )

        log_summary(model)
        ```
    """
    if wandb.run is None:
        raise wandb.Error("You must call `wandb.init()` before `log_summary()`")

    if not (isinstance(model, (CatBoostClassifier, CatBoostRegressor))):
        raise wandb.Error(
            "Model should be an instance of CatBoostClassifier or CatBoostRegressor"
        )

    with wb_telemetry.context() as tel:
        tel.feature.catboost_log_summary = True

    # log configs
    params = model.get_all_params()
    if log_all_params:
        wandb.config.update(params)

    # log best score and iteration
    wandb.run.summary["best_iteration"] = model.get_best_iteration()
    wandb.run.summary["best_score"] = model.get_best_score()

    # log model
    if save_model_checkpoint:
        aliases = ["best"] if params["use_best_model"] else ["last"]
        _checkpoint_artifact(model, aliases=aliases)

    # Feature importance
    if log_feature_importance:
        _log_feature_importance(model)



"""Implementation of Google Artifact Registry for wandb launch."""
import logging
import re
from typing import Tuple

from wandb.sdk.launch.environment.gcp_environment import GcpEnvironment
from wandb.sdk.launch.utils import LaunchError
from wandb.util import get_module

from .abstract import AbstractRegistry

google = get_module(
    "google",
    required="Google Cloud Platform support requires the google package. Please"
    " install it with `pip install wandb[launch]`.",
)

google.cloud.artifactregistry = get_module(
    "google.cloud.artifactregistry",
    required="Google Cloud Platform support requires the google-cloud-artifact-registry package. "
    "Please install it with `pip install wandb[launch]`.",
)

google.auth.credentials = get_module(
    "google.auth.credentials",
    required="Google Cloud Platform support requires google-auth. "
    "Please install it with `pip install wandb[launch]`.",
)

_logger = logging.getLogger(__name__)


class GoogleArtifactRegistry(AbstractRegistry):
    """Google Artifact Registry.

    Attributes:
        repository: The repository name.
        environment: A GcpEnvironment configured for access to this registry.
    """

    repository: str
    image_name: str
    environment: GcpEnvironment

    def __init__(
        self,
        repository: str,
        image_name: str,
        environment: GcpEnvironment,
        verify: bool = True,
    ) -> None:
        """Initialize the Google Artifact Registry.

        Arguments:
            repository: The repository name.
            image_name: The image name.
            environment: A GcpEnvironment configured for access to this registry.
            verify: Whether to verify the credentials, region, and project.

        Raises:
            LaunchError: If verify is True and the container registry or its
                environment have not been properly configured. Or if the environment
                is not an instance of GcpEnvironment.
        """
        _logger.info(
            f"Initializing Google Artifact Registry with repository {repository} "
            f"and image name {image_name}"
        )
        self.repository = repository
        self.image_name = image_name
        if not re.match(r"^\w[\w.-]+$", image_name):
            raise LaunchError(
                f"The image name {image_name} is invalid. The image name must "
                "consist of alphanumeric characters and underscores."
            )
        self.environment = environment
        if verify:
            self.verify()

    @property
    def uri(self) -> str:
        """The uri of the registry."""
        return f"{self.environment.region}-docker.pkg.dev/{self.environment.project}/{self.repository}/{self.image_name}"

    @uri.setter
    def uri(self, uri: str) -> None:
        """Set the uri of the registry."""
        raise LaunchError("The uri of the Google Artifact Registry cannot be set.")

    @classmethod
    def from_config(  # type: ignore[override]
        cls,
        config: dict,
        environment: GcpEnvironment,
        verify: bool = True,
    ) -> "GoogleArtifactRegistry":
        """Create a Google Artifact Registry from a config.

        Arguments:
            config: A dictionary containing the following keys:
                repository: The repository name.
                image_name: The image name.
            environment: A GcpEnvironment configured for access to this registry.

        Returns:
            A GoogleArtifactRegistry.
        """
        repository = config.get("repository")
        if not repository:
            raise LaunchError(
                "The Google Artifact Registry repository must be specified."
            )
        image_name = config.get("image_name")
        if not image_name:
            raise LaunchError("The image name must be specified.")
        return cls(repository, image_name, environment, verify=verify)

    def verify(self) -> None:
        """Verify the registry is properly configured.

        Raises:
            LaunchError: If the registry is not properly configured.
        """
        credentials = self.environment.get_credentials()
        parent = (
            f"projects/{self.environment.project}/locations/{self.environment.region}"
        )
        # We need to list the repositories to verify that the repository exists.
        request = google.cloud.artifactregistry.ListRepositoriesRequest(parent=parent)
        client = google.cloud.artifactregistry.ArtifactRegistryClient(
            credentials=credentials
        )
        try:
            response = client.list_repositories(request=request)
        except google.api_core.exceptions.PermissionDenied:
            raise LaunchError(
                "The provided credentials do not have permission to access the "
                f"Google Artifact Registry repository {self.repository}."
            )
        # Look for self.repository in the list of responses.
        for repo in response:
            if repo.name.endswith(self.repository):
                break
        # If we didn't find the repository, raise an error.
        else:
            raise LaunchError(
                f"The Google Artifact Registry repository {self.repository} does not exist."
            )

    def get_username_password(self) -> Tuple[str, str]:
        """Get the username and password for the registry.

        Returns:
            A tuple of the username and password.
        """
        credentials = self.environment.get_credentials()
        return "oauth2accesstoken", credentials.token

    def get_repo_uri(self) -> str:
        """Get the URI for the given repository.

        Arguments:
            repo_name: The repository name.

        Returns:
            The repository URI.
        """
        return (
            f"{self.environment.region}-docker.pkg.dev/"
            f"{self.environment.project}/{self.repository}/{self.image_name}"
        )

    def check_image_exists(self, image_uri: str) -> bool:
        """Check if the image exists.

        Arguments:
            image_uri: The image URI.

        Returns:
            True if the image exists, False otherwise.
        """
        _logger.info(
            f"Checking if image {image_uri} exists. In Google Artifact Registry {self.uri}."
        )

        return False
        # TODO: Test GCP Artifact Registry image exists to get working
        # repo_uri, _ = image_uri.split(":")
        # if repo_uri != self.get_repo_uri():
        #     raise LaunchError(
        #         f"The image {image_uri} does not belong to the Google Artifact "
        #         f"Repository {self.get_repo_uri()}."
        #     )
        # credentials = self.environment.get_credentials()
        # request = google.cloud.artifactregistry.GetTagRequest(parent=image_uri)
        # client = google.cloud.artifactregistry.ArtifactRegistryClient(
        #     credentials=credentials
        # )
        # try:
        #     client.get_tag(request=request)
        #     return True
        # except google.api_core.exceptions.NotFound:
        #     return False



from ...error import GraphQLError
from ...utils.quoted_or_list import quoted_or_list
from ...utils.suggestion_list import suggestion_list
from .base import ValidationRule


def _unknown_type_message(type, suggested_types):
    message = 'Unknown type "{}".'.format(type)
    if suggested_types:
        message += ' Perhaps you meant {}?'.format(quoted_or_list(suggested_types))

    return message


class KnownTypeNames(ValidationRule):

    def enter_ObjectTypeDefinition(self, node, *args):
        return False

    def enter_InterfaceTypeDefinition(self, node, *args):
        return False

    def enter_UnionTypeDefinition(self, node, *args):
        return False

    def enter_InputObjectTypeDefinition(self, node, *args):
        return False

    def enter_NamedType(self, node, *args):
        schema = self.context.get_schema()
        type_name = node.name.value
        type = schema.get_type(type_name)

        if not type:
            self.context.report_error(
                GraphQLError(
                    _unknown_type_message(
                        type_name,
                        suggestion_list(type_name, list(schema.get_type_map().keys()))
                    ),
                    [node]
                )
            )



"""Scheduler for classic wandb Sweeps."""
import logging
from pprint import pformat as pf
from typing import Any, Dict, List, Optional

import wandb
from wandb.sdk.launch.sweeps.scheduler import LOG_PREFIX, RunState, Scheduler, SweepRun

_logger = logging.getLogger(__name__)


class SweepScheduler(Scheduler):
    """A controller/agent that populates a Launch RunQueue from a sweeps RunQueue."""

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

    def _get_next_sweep_run(self, worker_id: int) -> Optional[SweepRun]:
        """Called by the main scheduler execution loop.

        Expected to return a properly formatted SweepRun if the scheduler
        is alive, or None and set the appropriate scheduler state:

        FAILED: self.fail_sweep()
        STOPPED: self.stop_sweep()
        """
        commands: List[Dict[str, Any]] = self._get_sweep_commands(worker_id)
        for command in commands:
            # The command "type" can be one of "run", "resume", "stop", "exit"
            _type = command.get("type")
            if _type in ["exit", "stop"]:
                self.stop_sweep()
                return None

            if _type not in ["run", "resume"]:
                self.fail_sweep(f"AgentHeartbeat unknown command: {_type}")

            _run_id: Optional[str] = command.get("run_id")
            if not _run_id:
                self.fail_sweep(f"No run id in agent heartbeat: {command}")
                return None

            if _run_id in self._runs:
                wandb.termlog(f"{LOG_PREFIX}Skipping duplicate run: {_run_id}")
                continue

            return SweepRun(
                id=_run_id,
                args=command.get("args", {}),
                logs=command.get("logs", []),
                worker_id=worker_id,
            )
        return None

    def _get_sweep_commands(self, worker_id: int) -> List[Dict[str, Any]]:
        """Helper to recieve sweep command from backend."""
        # AgentHeartbeat wants a Dict of runs which are running or queued
        _run_states: Dict[str, bool] = {}
        for run_id, run in self._yield_runs():
            # Filter out runs that are from a different worker thread
            if run.worker_id == worker_id and run.state == RunState.ALIVE:
                _run_states[run_id] = True

        _logger.debug(f"Sending states: \n{pf(_run_states)}\n")
        commands: List[Dict[str, Any]] = self._api.agent_heartbeat(
            agent_id=self._workers[worker_id].agent_id,
            metrics={},
            run_states=_run_states,
        )
        _logger.debug(f"AgentHeartbeat commands: \n{pf(commands)}\n")

        return commands

    def _exit(self) -> None:
        pass

    def _poll(self) -> None:
        _logger.debug(f"_poll. _runs: {self._runs}")
        pass

    def _load_state(self) -> None:
        pass

    def _save_state(self) -> None:
        pass



"""Common utilities for the LangChain integration.

This file exposes 4 primary functions:
- `print_wandb_init_message`: Prints a message to the user when the `WandbTracer` is initialized.
- `safely_convert_lc_run_to_wb_span`: Converts a LangChain Run into a W&B Trace Span.
- `safely_get_span_producing_model`: Retrieves the model that produced a given LangChain Run.
- `safely_convert_model_to_dict`: Converts a LangChain model into a dictionary.

These functions are used by the `WandbTracer` to extract and save the relevant information.
"""

from typing import TYPE_CHECKING, Any, Optional, Union

from langchain.callbacks.tracers.schemas import Run, RunTypeEnum

import wandb
from wandb.sdk.data_types import trace_tree

if TYPE_CHECKING:
    from langchain.chains.base import Chain
    from langchain.llms.base import BaseLLM
    from langchain.schema import BaseLanguageModel
    from langchain.tools.base import BaseTool


PRINT_WARNINGS = True


def print_wandb_init_message(run_url: str) -> None:
    wandb.termlog(
        f"Streaming LangChain activity to W&B at {run_url}\n"
        "`WandbTracer` is currently in beta.\n"
        "Please report any issues to https://github.com/wandb/wandb/issues with the tag `langchain`."
    )


def safely_convert_lc_run_to_wb_span(run: Run) -> Optional["trace_tree.Span"]:
    try:
        return _convert_lc_run_to_wb_span(run)
    except Exception as e:
        if PRINT_WARNINGS:
            wandb.termwarn(
                f"Skipping trace saving - unable to safely convert LangChain Run into W&B Trace due to: {e}"
            )
    return None


def safely_get_span_producing_model(run: Run) -> Any:
    try:
        return run.serialized.get("_self")
    except Exception as e:
        if PRINT_WARNINGS:
            wandb.termwarn(
                f"Skipping model saving - unable to safely retrieve LangChain model due to: {e}"
            )
    return None


def safely_convert_model_to_dict(
    model: Union["BaseLanguageModel", "BaseLLM", "BaseTool", "Chain"]
) -> Optional[dict]:
    """Returns the model dict if possible, otherwise returns None.

    Given that Models are all user defined, this operation is not always possible.
    """
    data = None
    message = None
    try:
        data = model.dict()
    except Exception as e:
        message = str(e)
        if hasattr(model, "agent"):
            try:
                data = model.agent.dict()
            except Exception as e:
                message = str(e)

    if data is not None and not isinstance(data, dict):
        message = (
            f"Model's dict transformation resulted in {type(data)}, expected a dict."
        )
        data = None

    if data is not None:
        data = _replace_type_with_kind(data)
    else:
        if PRINT_WARNINGS:
            wandb.termwarn(
                f"Skipping model saving - unable to safely convert LangChain Model to dictionary due to: {message}"
            )

    return data


def _convert_lc_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    if run.run_type == RunTypeEnum.llm:
        return _convert_llm_run_to_wb_span(run)
    elif run.run_type == RunTypeEnum.chain:
        return _convert_chain_run_to_wb_span(run)
    elif run.run_type == RunTypeEnum.tool:
        return _convert_tool_run_to_wb_span(run)
    else:
        return _convert_run_to_wb_span(run)


def _convert_llm_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    base_span = _convert_run_to_wb_span(run)

    base_span.results = [
        trace_tree.Result(
            inputs={"prompt": prompt},
            outputs={
                f"gen_{g_i}": gen["text"]
                for g_i, gen in enumerate(run.outputs["generations"][ndx])
            }
            if (
                run.outputs is not None
                and len(run.outputs["generations"]) > ndx
                and len(run.outputs["generations"][ndx]) > 0
            )
            else None,
        )
        for ndx, prompt in enumerate(run.inputs["prompts"] or [])
    ]
    base_span.span_kind = trace_tree.SpanKind.LLM

    return base_span


def _convert_chain_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    base_span = _convert_run_to_wb_span(run)

    base_span.results = [trace_tree.Result(inputs=run.inputs, outputs=run.outputs)]
    base_span.child_spans = [
        _convert_lc_run_to_wb_span(child_run) for child_run in run.child_runs
    ]
    base_span.span_kind = (
        trace_tree.SpanKind.AGENT
        if "agent" in run.serialized.get("name").lower()
        else trace_tree.SpanKind.CHAIN
    )

    return base_span


def _convert_tool_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    base_span = _convert_run_to_wb_span(run)

    base_span.attributes["input"] = run.inputs["input"]
    base_span.results = [trace_tree.Result(inputs=run.inputs, outputs=run.outputs)]
    base_span.child_spans = [
        _convert_lc_run_to_wb_span(child_run) for child_run in run.child_runs
    ]
    base_span.span_kind = trace_tree.SpanKind.TOOL

    return base_span


def _convert_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    attributes = {**run.extra} if run.extra else {}
    attributes["execution_order"] = run.execution_order

    return trace_tree.Span(
        span_id=str(run.id) if run.id is not None else None,
        name=run.serialized.get("name"),
        start_time_ms=int(run.start_time.timestamp() * 1000),
        end_time_ms=int(run.end_time.timestamp() * 1000),
        status_code=trace_tree.StatusCode.SUCCESS
        if run.error is None
        else trace_tree.StatusCode.ERROR,
        status_message=run.error,
        attributes=attributes,
    )


def _replace_type_with_kind(data: dict) -> dict:
    if isinstance(data, dict):
        # W&B TraceTree expects "_kind" instead of "_type" since `_type` is special
        # in W&B.
        if "_type" in data:
            _type = data.pop("_type")
            data["_kind"] = _type
        return {k: _replace_type_with_kind(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [_replace_type_with_kind(v) for v in data]
    elif isinstance(data, tuple):
        return tuple(_replace_type_with_kind(v) for v in data)
    elif isinstance(data, set):
        return {_replace_type_with_kind(v) for v in data}
    else:
        return data



"""A simple GraphQL client for sending queries and mutations.

Note: This was originally wandb/vendor/gql-0.2.0/wandb_gql/transport/requests.py
The only substantial change is to re-use a requests.Session object.
"""

from typing import Any, Callable, Dict, Optional, Tuple, Union

import requests
from wandb_gql.transport.http import HTTPTransport
from wandb_graphql.execution import ExecutionResult
from wandb_graphql.language import ast
from wandb_graphql.language.printer import print_ast


class GraphQLSession(HTTPTransport):
    def __init__(
        self,
        url: str,
        auth: Optional[Union[Tuple[str, str], Callable]] = None,
        use_json: bool = False,
        timeout: Optional[Union[int, float]] = None,
        **kwargs: Any,
    ) -> None:
        """Setup a session for sending GraphQL queries and mutations.

        Args:
            url (str): The GraphQL URL
            auth (tuple or callable): Auth tuple or callable for Basic/Digest/Custom HTTP Auth
            use_json (bool): Send request body as JSON instead of form-urlencoded
            timeout (int, float): Specifies a default timeout for requests (Default: None)
        """
        super().__init__(url, **kwargs)
        self.session = requests.Session()
        self.session.auth = auth
        self.default_timeout = timeout
        self.use_json = use_json

    def execute(
        self,
        document: ast.Node,
        variable_values: Optional[Dict] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> ExecutionResult:
        query_str = print_ast(document)
        payload = {"query": query_str, "variables": variable_values or {}}

        data_key = "json" if self.use_json else "data"
        post_args = {
            "headers": self.headers,
            "cookies": self.cookies,
            "timeout": timeout or self.default_timeout,
            data_key: payload,
        }
        request = self.session.post(self.url, **post_args)
        request.raise_for_status()

        result = request.json()
        data, errors = result.get("data"), result.get("errors")
        if data is None and errors is None:
            raise RuntimeError(f"Received non-compatible response: {result}")
        return ExecutionResult(data=data, errors=errors)



__all__ = ("Sentry",)


import atexit
import functools
import os
import pathlib
import sys
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, Type, Union
from urllib.parse import quote

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

import sentry_sdk  # type: ignore
import sentry_sdk.utils  # type: ignore

import wandb
import wandb.env
import wandb.util

if TYPE_CHECKING:
    import wandb.sdk.internal.settings_static

SENTRY_DEFAULT_DSN = (
    "https://2592b1968ea94cca9b5ef5e348e094a7@o151352.ingest.sentry.io/4504800232407040"
)

SessionStatus = Literal["ok", "exited", "crashed", "abnormal"]


def _noop_if_disabled(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(self: Type["Sentry"], *args: Any, **kwargs: Any) -> Any:
        if self._disabled:
            return None
        return func(self, *args, **kwargs)

    return wrapper


class Sentry:
    _disabled: bool

    def __init__(self) -> None:
        self._disabled = not wandb.env.error_reporting_enabled()
        self._sent_messages: set = set()

        self.dsn = os.environ.get(wandb.env.SENTRY_DSN, SENTRY_DEFAULT_DSN)

        self.hub: Optional["sentry_sdk.hub.Hub"] = None

        # ensure we always end the Sentry session
        atexit.register(self.end_session)

    @property
    def environment(self) -> str:
        """Return the environment we're running in."""
        # check if we're in a git repo
        is_git = pathlib.Path(__file__).parent.parent.parent.joinpath(".git").exists()

        # these match the environments for gorilla
        return "development" if is_git else "production"

    @_noop_if_disabled
    def setup(self) -> None:
        """Setup Sentry SDK.

        We use lower-level APIs (i.e., not sentry_sdk.init) here
        to avoid the possibility of interfering with the user's
        own Sentry SDK setup.
        """
        client = sentry_sdk.Client(
            dsn=self.dsn,
            default_integrations=False,
            environment=self.environment,
            release=wandb.__version__,
        )
        self.hub = sentry_sdk.Hub(client)

    @_noop_if_disabled
    def message(self, message: str, repeat: bool = True) -> None:
        """Send a message to Sentry."""
        if not repeat and message in self._sent_messages:
            return
        self._sent_messages.add(message)
        self.hub.capture_message(message)  # type: ignore

    @_noop_if_disabled
    def exception(
        self,
        exc: Union[
            str,
            BaseException,
            Tuple[
                Optional[Type[BaseException]],
                Optional[BaseException],
                Optional[TracebackType],
            ],
            None,
        ],
        handled: bool = False,
        status: Optional["SessionStatus"] = None,
    ) -> None:
        """Log an exception to Sentry."""
        error = Exception(exc) if isinstance(exc, str) else exc
        # based on self.hub.capture_exception(_exc)
        if error is not None:
            exc_info = sentry_sdk.utils.exc_info_from_error(error)
        else:
            exc_info = sys.exc_info()

        event, hint = sentry_sdk.utils.event_from_exception(
            exc_info,
            client_options=self.hub.client.options,  # type: ignore
            mechanism={"type": "generic", "handled": handled},
        )
        try:
            self.hub.capture_event(event, hint=hint)  # type: ignore
        except Exception:
            pass

        # if the status is not explicitly set, we'll set it to "crashed" if the exception
        # was unhandled, or "errored" if it was handled
        status = status or ("crashed" if not handled else "errored")  # type: ignore
        self.mark_session(status=status)

        client, _ = self.hub._stack[-1]
        client.flush()

        return None

    def reraise(self, exc: Any) -> None:
        """Re-raise an exception after logging it to Sentry.

        Use this for top-level exceptions when you want the user to see the traceback.

        Must be called from within an exception handler.
        """
        self.exception(exc)
        # this will messily add this "reraise" function to the stack trace,
        # but hopefully it's not too bad
        raise exc.with_traceback(sys.exc_info()[2])

    @_noop_if_disabled
    def start_session(self) -> None:
        """Start a new session."""
        assert self.hub is not None
        # get the current client and scope
        _, scope = self.hub._stack[-1]
        session = scope._session

        # if there's no session, start one
        if session is None:
            self.hub.start_session()

    @_noop_if_disabled
    def end_session(self) -> None:
        """End the current session."""
        assert self.hub is not None
        # get the current client and scope
        client, scope = self.hub._stack[-1]
        session = scope._session

        if session is not None and client is not None:
            self.hub.end_session()
            client.flush()

    @_noop_if_disabled
    def mark_session(self, status: Optional["SessionStatus"] = None) -> None:
        """Mark the current session with a status."""
        assert self.hub is not None
        _, scope = self.hub._stack[-1]
        session = scope._session

        if session is not None:
            session.update(status=status)

    @_noop_if_disabled
    def configure_scope(
        self,
        settings: Optional[
            Union[
                "wandb.sdk.wandb_settings.Settings",
                "wandb.sdk.internal.settings_static.SettingsStatic",
            ]
        ] = None,
        process_context: Optional[str] = None,
    ) -> None:
        """Configure the Sentry scope for the current thread.

        This function should be called at the beginning of every thread that
        will send events to Sentry. It sets the tags that will be applied to
        all events sent from this thread. It also tries to start a session
        if one doesn't already exist for this thread.
        """
        assert self.hub is not None
        settings_tags = (
            "entity",
            "project",
            "run_id",
            "run_url",
            "sweep_url",
            "sweep_id",
            "deployment",
            "_disable_service",
            "launch",
        )

        with self.hub.configure_scope() as scope:
            scope.set_tag("platform", wandb.util.get_platform_name())

            # set context
            if process_context:
                scope.set_tag("process_context", process_context)

            # apply settings tags
            if settings is None:
                return None

            for tag in settings_tags:
                val = getattr(settings, tag, None)
                if val not in (None, ""):
                    scope.set_tag(tag, val)

            # todo: update once #4982 is merged
            python_runtime = (
                "colab"
                if getattr(settings, "_colab", None)
                else ("jupyter" if getattr(settings, "_jupyter", None) else "python")
            )
            scope.set_tag("python_runtime", python_runtime)

            # Hack for constructing run_url and sweep_url given run_id and sweep_id
            required = ("entity", "project", "base_url")
            params = {key: getattr(settings, key, None) for key in required}
            if all(params.values()):
                # here we're guaranteed that entity, project, base_url all have valid values
                app_url = wandb.util.app_url(params["base_url"])  # type: ignore
                ent, proj = (quote(params[k]) for k in ("entity", "project"))  # type: ignore

                # TODO: the settings object will be updated to contain run_url and sweep_url
                # This is done by passing a settings_map in the run_start protocol buffer message
                for word in ("run", "sweep"):
                    _url, _id = f"{word}_url", f"{word}_id"
                    if not getattr(settings, _url, None) and getattr(
                        settings, _id, None
                    ):
                        scope.set_tag(
                            _url,
                            f"{app_url}/{ent}/{proj}/{word}s/{getattr(settings, _id, None)}",
                        )

            if hasattr(settings, "email"):
                scope.user = {"email": settings.email}  # noqa

        # todo: add back the option to pass general tags see: c645f625d1c1a3db4a6b0e2aa8e924fee101904c (wandb/util.py)

        self.start_session()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: wandb/proto/wandb_telemetry.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from wandb.proto import wandb_base_pb2 as wandb_dot_proto_dot_wandb__base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!wandb/proto/wandb_telemetry.proto\x12\x0ewandb_internal\x1a\x1cwandb/proto/wandb_base.proto\"\xb3\x03\n\x0fTelemetryRecord\x12-\n\x0cimports_init\x18\x01 \x01(\x0b\x32\x17.wandb_internal.Imports\x12/\n\x0eimports_finish\x18\x02 \x01(\x0b\x32\x17.wandb_internal.Imports\x12(\n\x07\x66\x65\x61ture\x18\x03 \x01(\x0b\x32\x17.wandb_internal.Feature\x12\x16\n\x0epython_version\x18\x04 \x01(\t\x12\x13\n\x0b\x63li_version\x18\x05 \x01(\t\x12\x1b\n\x13huggingface_version\x18\x06 \x01(\t\x12 \n\x03\x65nv\x18\x08 \x01(\x0b\x32\x13.wandb_internal.Env\x12%\n\x05label\x18\t \x01(\x0b\x32\x16.wandb_internal.Labels\x12.\n\ndeprecated\x18\n \x01(\x0b\x32\x1a.wandb_internal.Deprecated\x12&\n\x06issues\x18\x0b \x01(\x0b\x32\x16.wandb_internal.Issues\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\"\x11\n\x0fTelemetryResult\"\x9f\x0b\n\x07Imports\x12\r\n\x05torch\x18\x01 \x01(\x08\x12\r\n\x05keras\x18\x02 \x01(\x08\x12\x12\n\ntensorflow\x18\x03 \x01(\x08\x12\x0e\n\x06\x66\x61stai\x18\x04 \x01(\x08\x12\x0f\n\x07sklearn\x18\x05 \x01(\x08\x12\x0f\n\x07xgboost\x18\x06 \x01(\x08\x12\x10\n\x08\x63\x61tboost\x18\x07 \x01(\x08\x12\x10\n\x08lightgbm\x18\x08 \x01(\x08\x12\x19\n\x11pytorch_lightning\x18\t \x01(\x08\x12\x0e\n\x06ignite\x18\n \x01(\x08\x12\x14\n\x0ctransformers\x18\x0b \x01(\x08\x12\x0b\n\x03jax\x18\x0c \x01(\x08\x12\x10\n\x08metaflow\x18\r \x01(\x08\x12\x10\n\x08\x61llennlp\x18\x0e \x01(\x08\x12\x11\n\tautogluon\x18\x0f \x01(\x08\x12\x11\n\tautokeras\x18\x10 \x01(\x08\x12\x10\n\x08\x63\x61talyst\x18\x12 \x01(\x08\x12\x10\n\x08\x64\x65\x65pchem\x18\x15 \x01(\x08\x12\x0f\n\x07\x64\x65\x65pctr\x18\x16 \x01(\x08\x12\x0f\n\x07pycaret\x18\x1c \x01(\x08\x12\x14\n\x0cpytorchvideo\x18\x1d \x01(\x08\x12\x0b\n\x03ray\x18\x1e \x01(\x08\x12\x1a\n\x12simpletransformers\x18\x1f \x01(\x08\x12\x0e\n\x06skorch\x18  \x01(\x08\x12\r\n\x05spacy\x18! \x01(\x08\x12\r\n\x05\x66lash\x18\" \x01(\x08\x12\x0e\n\x06optuna\x18# \x01(\x08\x12\x0f\n\x07recbole\x18$ \x01(\x08\x12\x0c\n\x04mmcv\x18% \x01(\x08\x12\r\n\x05mmdet\x18& \x01(\x08\x12\x11\n\ttorchdrug\x18\' \x01(\x08\x12\x11\n\ttorchtext\x18( \x01(\x08\x12\x13\n\x0btorchvision\x18) \x01(\x08\x12\r\n\x05\x65legy\x18* \x01(\x08\x12\x12\n\ndetectron2\x18+ \x01(\x08\x12\r\n\x05\x66lair\x18, \x01(\x08\x12\x0c\n\x04\x66lax\x18- \x01(\x08\x12\x0c\n\x04syft\x18. \x01(\x08\x12\x0b\n\x03TTS\x18/ \x01(\x08\x12\r\n\x05monai\x18\x30 \x01(\x08\x12\x17\n\x0fhuggingface_hub\x18\x31 \x01(\x08\x12\r\n\x05hydra\x18\x32 \x01(\x08\x12\x10\n\x08\x64\x61tasets\x18\x33 \x01(\x08\x12\x0e\n\x06sacred\x18\x34 \x01(\x08\x12\x0e\n\x06joblib\x18\x35 \x01(\x08\x12\x0c\n\x04\x64\x61sk\x18\x36 \x01(\x08\x12\x0f\n\x07\x61syncio\x18\x37 \x01(\x08\x12\x11\n\tpaddleocr\x18\x38 \x01(\x08\x12\r\n\x05ppdet\x18\x39 \x01(\x08\x12\x11\n\tpaddleseg\x18: \x01(\x08\x12\x11\n\tpaddlenlp\x18; \x01(\x08\x12\r\n\x05mmseg\x18< \x01(\x08\x12\r\n\x05mmocr\x18= \x01(\x08\x12\r\n\x05mmcls\x18> \x01(\x08\x12\x0c\n\x04timm\x18? \x01(\x08\x12\x0f\n\x07\x66\x61irseq\x18@ \x01(\x08\x12\x12\n\ndeepchecks\x18\x41 \x01(\x08\x12\x10\n\x08\x63omposer\x18\x42 \x01(\x08\x12\x10\n\x08sparseml\x18\x43 \x01(\x08\x12\x10\n\x08\x61nomalib\x18\x44 \x01(\x08\x12\r\n\x05zenml\x18\x45 \x01(\x08\x12\x12\n\ncolossalai\x18\x46 \x01(\x08\x12\x12\n\naccelerate\x18G \x01(\x08\x12\x0e\n\x06merlin\x18H \x01(\x08\x12\x0f\n\x07nanodet\x18I \x01(\x08\x12#\n\x1bsegmentation_models_pytorch\x18J \x01(\x08\x12\x1d\n\x15sentence_transformers\x18K \x01(\x08\x12\x0b\n\x03\x64gl\x18L \x01(\x08\x12\x17\n\x0ftorch_geometric\x18M \x01(\x08\x12\x0c\n\x04jina\x18N \x01(\x08\x12\x0e\n\x06kornia\x18O \x01(\x08\x12\x16\n\x0e\x61lbumentations\x18P \x01(\x08\x12\x10\n\x08keras_cv\x18Q \x01(\x08\x12\x10\n\x08mmengine\x18R \x01(\x08\x12\x11\n\tdiffusers\x18S \x01(\x08\x12\x0b\n\x03trl\x18T \x01(\x08\x12\x0c\n\x04trlx\x18U \x01(\x08\x12\x11\n\tlangchain\x18V \x01(\x08\x12\x13\n\x0bllama_index\x18W \x01(\x08\x12\x15\n\rstability_sdk\x18X \x01(\x08\"\xda\t\n\x07\x46\x65\x61ture\x12\r\n\x05watch\x18\x01 \x01(\x08\x12\x0e\n\x06\x66inish\x18\x02 \x01(\x08\x12\x0c\n\x04save\x18\x03 \x01(\x08\x12\x0f\n\x07offline\x18\x04 \x01(\x08\x12\x0f\n\x07resumed\x18\x05 \x01(\x08\x12\x0c\n\x04grpc\x18\x06 \x01(\x08\x12\x0e\n\x06metric\x18\x07 \x01(\x08\x12\r\n\x05keras\x18\x08 \x01(\x08\x12\x11\n\tsagemaker\x18\t \x01(\x08\x12\x1c\n\x14\x61rtifact_incremental\x18\n \x01(\x08\x12\x10\n\x08metaflow\x18\x0b \x01(\x08\x12\x0f\n\x07prodigy\x18\x0c \x01(\x08\x12\x15\n\rset_init_name\x18\r \x01(\x08\x12\x13\n\x0bset_init_id\x18\x0e \x01(\x08\x12\x15\n\rset_init_tags\x18\x0f \x01(\x08\x12\x17\n\x0fset_init_config\x18\x10 \x01(\x08\x12\x14\n\x0cset_run_name\x18\x11 \x01(\x08\x12\x14\n\x0cset_run_tags\x18\x12 \x01(\x08\x12\x17\n\x0fset_config_item\x18\x13 \x01(\x08\x12\x0e\n\x06launch\x18\x14 \x01(\x08\x12\x1c\n\x14torch_profiler_trace\x18\x15 \x01(\x08\x12\x0b\n\x03sb3\x18\x16 \x01(\x08\x12\x0f\n\x07service\x18\x17 \x01(\x08\x12\x17\n\x0finit_return_run\x18\x18 \x01(\x08\x12\x1f\n\x17lightgbm_wandb_callback\x18\x19 \x01(\x08\x12\x1c\n\x14lightgbm_log_summary\x18\x1a \x01(\x08\x12\x1f\n\x17\x63\x61tboost_wandb_callback\x18\x1b \x01(\x08\x12\x1c\n\x14\x63\x61tboost_log_summary\x18\x1c \x01(\x08\x12\x17\n\x0ftensorboard_log\x18\x1d \x01(\x08\x12\x16\n\x0e\x65stimator_hook\x18\x1e \x01(\x08\x12\x1e\n\x16xgboost_wandb_callback\x18\x1f \x01(\x08\x12\"\n\x1axgboost_old_wandb_callback\x18  \x01(\x08\x12\x0e\n\x06\x61ttach\x18! \x01(\x08\x12\x19\n\x11tensorboard_patch\x18\" \x01(\x08\x12\x18\n\x10tensorboard_sync\x18# \x01(\x08\x12\x15\n\rkfp_wandb_log\x18$ \x01(\x08\x12\x1b\n\x13maybe_run_overwrite\x18% \x01(\x08\x12\x1c\n\x14keras_metrics_logger\x18& \x01(\x08\x12\x1e\n\x16keras_model_checkpoint\x18\' \x01(\x08\x12!\n\x19keras_wandb_eval_callback\x18( \x01(\x08\x12\x1d\n\x15\x66low_control_overflow\x18) \x01(\x08\x12\x0c\n\x04sync\x18* \x01(\x08\x12\x1d\n\x15\x66low_control_disabled\x18+ \x01(\x08\x12\x1b\n\x13\x66low_control_custom\x18, \x01(\x08\x12\x18\n\x10service_disabled\x18- \x01(\x08\x12\x14\n\x0copen_metrics\x18. \x01(\x08\x12\x1a\n\x12ultralytics_yolov8\x18/ \x01(\x08\x12\x17\n\x0fimporter_mlflow\x18\x30 \x01(\x08\x12\x15\n\rsync_tfevents\x18\x31 \x01(\x08\x12\x15\n\rasync_uploads\x18\x32 \x01(\x08\x12\x16\n\x0eopenai_autolog\x18\x33 \x01(\x08\x12\x18\n\x10langchain_tracer\x18\x34 \x01(\x08\"\xf1\x01\n\x03\x45nv\x12\x0f\n\x07jupyter\x18\x01 \x01(\x08\x12\x0e\n\x06kaggle\x18\x02 \x01(\x08\x12\x0f\n\x07windows\x18\x03 \x01(\x08\x12\x0e\n\x06m1_gpu\x18\x04 \x01(\x08\x12\x13\n\x0bstart_spawn\x18\x05 \x01(\x08\x12\x12\n\nstart_fork\x18\x06 \x01(\x08\x12\x18\n\x10start_forkserver\x18\x07 \x01(\x08\x12\x14\n\x0cstart_thread\x18\x08 \x01(\x08\x12\x10\n\x08maybe_mp\x18\t \x01(\x08\x12\x10\n\x08trainium\x18\n \x01(\x08\x12\x0b\n\x03pex\x18\x0b \x01(\x08\x12\r\n\x05\x63olab\x18\x0c \x01(\x08\x12\x0f\n\x07ipython\x18\r \x01(\x08\"H\n\x06Labels\x12\x13\n\x0b\x63ode_string\x18\x01 \x01(\t\x12\x13\n\x0brepo_string\x18\x02 \x01(\t\x12\x14\n\x0c\x63ode_version\x18\x03 \x01(\t\"\x80\x02\n\nDeprecated\x12!\n\x19keras_callback__data_type\x18\x01 \x01(\x08\x12\x11\n\trun__mode\x18\x02 \x01(\x08\x12\x19\n\x11run__save_no_args\x18\x03 \x01(\x08\x12\x11\n\trun__join\x18\x04 \x01(\x08\x12\r\n\x05plots\x18\x05 \x01(\x08\x12\x15\n\rrun__log_sync\x18\x06 \x01(\x08\x12!\n\x19init__config_include_keys\x18\x07 \x01(\x08\x12!\n\x19init__config_exclude_keys\x18\x08 \x01(\x08\x12\"\n\x1akeras_callback__save_model\x18\t \x01(\x08\"|\n\x06Issues\x12%\n\x1dsettings__validation_warnings\x18\x01 \x01(\x08\x12!\n\x19settings__unexpected_args\x18\x02 \x01(\x08\x12(\n settings__preprocessing_warnings\x18\x03 \x01(\x08\x62\x06proto3')



_TELEMETRYRECORD = DESCRIPTOR.message_types_by_name['TelemetryRecord']
_TELEMETRYRESULT = DESCRIPTOR.message_types_by_name['TelemetryResult']
_IMPORTS = DESCRIPTOR.message_types_by_name['Imports']
_FEATURE = DESCRIPTOR.message_types_by_name['Feature']
_ENV = DESCRIPTOR.message_types_by_name['Env']
_LABELS = DESCRIPTOR.message_types_by_name['Labels']
_DEPRECATED = DESCRIPTOR.message_types_by_name['Deprecated']
_ISSUES = DESCRIPTOR.message_types_by_name['Issues']
TelemetryRecord = _reflection.GeneratedProtocolMessageType('TelemetryRecord', (_message.Message,), {
  'DESCRIPTOR' : _TELEMETRYRECORD,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.TelemetryRecord)
  })
_sym_db.RegisterMessage(TelemetryRecord)

TelemetryResult = _reflection.GeneratedProtocolMessageType('TelemetryResult', (_message.Message,), {
  'DESCRIPTOR' : _TELEMETRYRESULT,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.TelemetryResult)
  })
_sym_db.RegisterMessage(TelemetryResult)

Imports = _reflection.GeneratedProtocolMessageType('Imports', (_message.Message,), {
  'DESCRIPTOR' : _IMPORTS,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.Imports)
  })
_sym_db.RegisterMessage(Imports)

Feature = _reflection.GeneratedProtocolMessageType('Feature', (_message.Message,), {
  'DESCRIPTOR' : _FEATURE,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.Feature)
  })
_sym_db.RegisterMessage(Feature)

Env = _reflection.GeneratedProtocolMessageType('Env', (_message.Message,), {
  'DESCRIPTOR' : _ENV,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.Env)
  })
_sym_db.RegisterMessage(Env)

Labels = _reflection.GeneratedProtocolMessageType('Labels', (_message.Message,), {
  'DESCRIPTOR' : _LABELS,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.Labels)
  })
_sym_db.RegisterMessage(Labels)

Deprecated = _reflection.GeneratedProtocolMessageType('Deprecated', (_message.Message,), {
  'DESCRIPTOR' : _DEPRECATED,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.Deprecated)
  })
_sym_db.RegisterMessage(Deprecated)

Issues = _reflection.GeneratedProtocolMessageType('Issues', (_message.Message,), {
  'DESCRIPTOR' : _ISSUES,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.Issues)
  })
_sym_db.RegisterMessage(Issues)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TELEMETRYRECORD._serialized_start=84
  _TELEMETRYRECORD._serialized_end=519
  _TELEMETRYRESULT._serialized_start=521
  _TELEMETRYRESULT._serialized_end=538
  _IMPORTS._serialized_start=541
  _IMPORTS._serialized_end=1980
  _FEATURE._serialized_start=1983
  _FEATURE._serialized_end=3225
  _ENV._serialized_start=3228
  _ENV._serialized_end=3469
  _LABELS._serialized_start=3471
  _LABELS._serialized_end=3543
  _DEPRECATED._serialized_start=3546
  _DEPRECATED._serialized_end=3802
  _ISSUES._serialized_start=3804
  _ISSUES._serialized_end=3928
# @@protoc_insertion_point(module_scope)



# Configuration for pytest to automatically collect types.
# Thanks to Guilherme Salgado.
import pytest

try:
    import pyannotate_runtime
    PYANOTATE_PRESENT = True
except ImportError:
    PYANOTATE_PRESENT = False

if PYANOTATE_PRESENT:
    def pytest_collection_finish(session):
        """Handle the pytest collection finish hook: configure pyannotate.
        Explicitly delay importing `collect_types` until all tests have
        been collected.  This gives gevent a chance to monkey patch the
        world before importing pyannotate.
        """
        from pyannotate_runtime import collect_types
        collect_types.init_types_collection()

    @pytest.fixture(autouse=True)
    def collect_types_fixture():
        from pyannotate_runtime import collect_types
        collect_types.resume()
        yield
        collect_types.pause()

    def pytest_sessionfinish(session, exitstatus):
        from pyannotate_runtime import collect_types
        collect_types.dump_stats("type_info.json")



"""Abstract plugin class defining the interface needed to build container images for W&B Launch."""
from abc import ABC, abstractmethod
from typing import Any, Dict

from wandb.sdk.launch.environment.abstract import AbstractEnvironment
from wandb.sdk.launch.registry.abstract import AbstractRegistry

from .._project_spec import EntryPoint, LaunchProject


class AbstractBuilder(ABC):
    """Abstract plugin class defining the interface needed to build container images for W&B Launch."""

    builder_type: str
    environment: AbstractEnvironment
    registry: AbstractRegistry
    builder_config: Dict[str, Any]

    @abstractmethod
    def __init__(
        self,
        environment: AbstractEnvironment,
        registry: AbstractRegistry,
        verify: bool = True,
    ) -> None:
        """Initialize a builder.

        Arguments:
            builder_config: The builder config.
            registry: The registry to use.
            verify: Whether to verify the functionality of the builder.

        Raises:
            LaunchError: If the builder cannot be intialized or verified.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_config(
        cls,
        config: dict,
        environment: AbstractEnvironment,
        registry: AbstractRegistry,
        verify: bool = True,
    ) -> "AbstractBuilder":
        """Create a builder from a config dictionary.

        Arguments:
            config: The config dictionary.
            environment: The environment to use.
            registry: The registry to use.
            verify: Whether to verify the functionality of the builder.
            login: Whether to login to the registry immediately.

        Returns:
            The builder.
        """
        raise NotImplementedError

    @abstractmethod
    def build_image(
        self,
        launch_project: LaunchProject,
        entrypoint: EntryPoint,
    ) -> str:
        """Build the image for the given project.

        Arguments:
            launch_project: The project to build.
            build_ctx_path: The path to the build context.

        Returns:
            The image name.
        """
        raise NotImplementedError

    @abstractmethod
    def verify(self) -> None:
        """Verify that the builder can be used to build images.

        Raises:
            LaunchError: If the builder cannot be used to build images.
        """
        raise NotImplementedError



"""wandb server.

Start up grpc or socket transport servers.
"""

import logging
import os
import sys
from concurrent import futures
from typing import Optional

import wandb

from ..lib import tracelog
from . import _startup_debug, port_file
from .server_sock import SocketServer
from .streams import StreamMux


class WandbServer:
    _pid: Optional[int]
    _grpc_port: Optional[int]
    _sock_port: Optional[int]
    _debug: bool
    _serve_grpc: bool
    _serve_sock: bool
    _sock_server: Optional[SocketServer]
    _startup_debug_enabled: bool

    def __init__(
        self,
        grpc_port: Optional[int] = None,
        sock_port: Optional[int] = None,
        port_fname: Optional[str] = None,
        address: Optional[str] = None,
        pid: Optional[int] = None,
        debug: bool = True,
        serve_grpc: bool = False,
        serve_sock: bool = False,
    ) -> None:
        self._grpc_port = grpc_port
        self._sock_port = sock_port
        self._port_fname = port_fname
        self._address = address
        self._pid = pid
        self._debug = debug
        self._serve_grpc = serve_grpc
        self._serve_sock = serve_sock
        self._sock_server = None
        self._startup_debug_enabled = _startup_debug.is_enabled()

        if grpc_port:
            _ = wandb.util.get_module(
                "grpc",
                required="grpc port requires the grpcio library, run pip install wandb[grpc]",
            )

        if debug:
            logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

    def _inform_used_ports(
        self, grpc_port: Optional[int], sock_port: Optional[int]
    ) -> None:
        if not self._port_fname:
            return
        pf = port_file.PortFile(grpc_port=grpc_port, sock_port=sock_port)
        pf.write(self._port_fname)

    def _start_grpc(self, mux: StreamMux) -> int:
        import grpc

        from wandb.proto import wandb_server_pb2_grpc as spb_grpc

        from .server_grpc import WandbServicer

        address: str = self._address or "127.0.0.1"
        port: int = self._grpc_port or 0
        pid: int = self._pid or 0
        server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=10, thread_name_prefix="GrpcPoolThr")
        )
        servicer = WandbServicer(server=server, mux=mux)
        try:
            spb_grpc.add_InternalServiceServicer_to_server(servicer, server)
            port = server.add_insecure_port(f"{address}:{port}")
            mux.set_port(port)
            mux.set_pid(pid)
            server.start()
        except KeyboardInterrupt:
            mux.cleanup()
            server.stop(0)
            raise
        except Exception:
            mux.cleanup()
            server.stop(0)
            raise
        return port

    def _start_sock(self, mux: StreamMux) -> int:
        address: str = self._address or "127.0.0.1"
        port: int = self._sock_port or 0
        self._sock_server = SocketServer(mux=mux, address=address, port=port)
        try:
            self._sock_server.start()
            port = self._sock_server.port
            if self._pid:
                mux.set_pid(self._pid)
        except KeyboardInterrupt:
            mux.cleanup()
            raise
        except Exception:
            mux.cleanup()
            raise
        return port

    def _stop_servers(self) -> None:
        if self._sock_server:
            self._sock_server.stop()

    def _setup_tracelog(self) -> None:
        # TODO: remove this temporary hack, need to find a better way to pass settings
        # to the server.  for now lets just look at the environment variable we need
        tracelog_mode = os.environ.get("WANDB_TRACELOG")
        if tracelog_mode:
            tracelog.enable(tracelog_mode)

    def _startup_debug_print(self, message: str) -> None:
        if not self._startup_debug_enabled:
            return
        _startup_debug.print_message(message)

    def serve(self) -> None:
        self._setup_tracelog()
        mux = StreamMux()
        self._startup_debug_print("before_network")
        grpc_port = self._start_grpc(mux=mux) if self._serve_grpc else None
        sock_port = self._start_sock(mux=mux) if self._serve_sock else None
        self._startup_debug_print("after_network")
        self._inform_used_ports(grpc_port=grpc_port, sock_port=sock_port)
        self._startup_debug_print("after_inform")
        setproctitle = wandb.util.get_optional_module("setproctitle")
        if setproctitle:
            service_ver = 2
            pid = str(self._pid or 0)
            transport = "s" if sock_port else "g"
            port = grpc_port or sock_port or 0
            # this format is similar to wandb_manager token but it purely informative now
            # (consider unifying this in the future)
            service_id = f"{service_ver}-{pid}-{transport}-{port}"
            proc_title = f"wandb-service({service_id})"
            setproctitle.setproctitle(proc_title)
        self._startup_debug_print("before_loop")
        mux.loop()
        self._stop_servers()



import threading
from typing import MutableMapping, NamedTuple

import wandb


class FileStats(NamedTuple):
    deduped: bool
    total: int
    uploaded: int
    failed: bool
    artifact_file: bool


class Summary(NamedTuple):
    uploaded_bytes: int
    total_bytes: int
    deduped_bytes: int


class FileCountsByCategory(NamedTuple):
    artifact: int
    wandb: int
    media: int
    other: int


class Stats:
    def __init__(self) -> None:
        self._stats: MutableMapping[str, "FileStats"] = {}
        self._lock = threading.Lock()

    def init_file(
        self, save_name: str, size: int, is_artifact_file: bool = False
    ) -> None:
        with self._lock:
            self._stats[save_name] = FileStats(
                deduped=False,
                total=size,
                uploaded=0,
                failed=False,
                artifact_file=is_artifact_file,
            )

    def set_file_deduped(self, save_name: str) -> None:
        with self._lock:
            orig = self._stats[save_name]
            self._stats[save_name] = orig._replace(
                deduped=True,
                uploaded=orig.total,
            )

    def update_uploaded_file(self, save_name: str, total_uploaded: int) -> None:
        with self._lock:
            self._stats[save_name] = self._stats[save_name]._replace(
                uploaded=total_uploaded,
            )

    def update_failed_file(self, save_name: str) -> None:
        with self._lock:
            self._stats[save_name] = self._stats[save_name]._replace(
                uploaded=0,
                failed=True,
            )

    def summary(self) -> Summary:
        # Need to use list to ensure we get a copy, since other threads may
        # modify this while we iterate
        with self._lock:
            stats = list(self._stats.values())
        return Summary(
            uploaded_bytes=sum(f.uploaded for f in stats),
            total_bytes=sum(f.total for f in stats),
            deduped_bytes=sum(f.total for f in stats if f.deduped),
        )

    def file_counts_by_category(self) -> FileCountsByCategory:
        artifact_files = 0
        wandb_files = 0
        media_files = 0
        other_files = 0
        # Need to use list to ensure we get a copy, since other threads may
        # modify this while we iterate
        with self._lock:
            file_stats = list(self._stats.items())
        for save_name, stats in file_stats:
            if stats.artifact_file:
                artifact_files += 1
            elif wandb.wandb_lib.filenames.is_wandb_file(save_name):  # type: ignore[attr-defined]  # TODO(spencerpearson): this is probably synonymous with wandb.sdk.lib.filenames...?
                wandb_files += 1
            elif save_name.startswith("media"):
                media_files += 1
            else:
                other_files += 1
        return FileCountsByCategory(
            artifact=artifact_files,
            wandb=wandb_files,
            media=media_files,
            other=other_files,
        )
